#ifndef HISTORY_HPP
#define HISTORY_HPP

/////////////////////// Qt includes
#include <QList>
#include <QDateTime>
#include <QMap>


/////////////////////// Local includes

namespace msXpertSuite
{


	enum IntegrationType
	{
		NOT_SET = 0x0000,

		DATA_TO_RT = 1 << 0,
		DATA_TO_DT_MZ_DT_RANGE = 1 << 2,
		DATA_TO_DT_MZ_MZ_RANGE = 1 << 3,

		RT_TO_MZ = 1 << 4,
		RT_TO_DT = 1 << 5,

		MZ_TO_RT = 1 << 6,
		MZ_TO_DT = 1 << 7,

		DT_TO_RT = 1 << 8,
		DT_TO_MZ = 1 << 9,

		DATA_TO_DT_MZ_TO_RT_DT_RANGE = 1 << 10,
		DATA_TO_DT_MZ_TO_RT_MZ_RANGE = 1 << 11,

		DATA_TO_DT_MZ_TO_MZ_DT_RANGE = 1 << 12,
		DATA_TO_DT_MZ_TO_MZ_MZ_RANGE = 1 << 13,

		DATA_TO_DT_MZ_TO_DT_DT_RANGE = 1 << 14,
		DATA_TO_DT_MZ_TO_DT_MZ_RANGE = 1 << 15,

		// These ones should not be exposed as they do not go in
		// history items:
		//
		// Integrate from TIC chrom range to TIC intensity single value
		RT_TO_TIC_INT = 1 << 16,

		// Integrate from mass spectrum range to TIC intensity single value
		MZ_TO_TIC_INT = 1 << 17,

		// Integrate from drift spectrum range to TIC intensity single value
		DT_TO_TIC_INT = 1 << 18,

		// Integrate from colormap DT_MZ to TIC intensity single value (range data
		// for DT axis)
		DATA_TO_DT_MZ_TO_TIC_INT_DT_RANGE = 1 << 19,

		// Integrate from colormap DT_MZ to TIC intensity single value (range data
		// for MZ axis)
		DATA_TO_DT_MZ_TO_TIC_INT_MZ_RANGE = 1 << 20,

		DATA_TO_DT_MZ_TO_TIC_INT = (DATA_TO_DT_MZ_TO_TIC_INT_DT_RANGE | DATA_TO_DT_MZ_TO_TIC_INT_MZ_RANGE),

		RT_TO_ANY = (RT_TO_MZ | RT_TO_DT | RT_TO_TIC_INT),
		MZ_TO_ANY = (MZ_TO_RT | MZ_TO_DT | DATA_TO_DT_MZ_MZ_RANGE | DATA_TO_DT_MZ_TO_TIC_INT_MZ_RANGE | MZ_TO_TIC_INT),
		DT_TO_ANY = (DT_TO_RT | DT_TO_MZ | DATA_TO_DT_MZ_DT_RANGE | DATA_TO_DT_MZ_TO_TIC_INT_DT_RANGE | DT_TO_TIC_INT),

		ANY_TO_RT = (MZ_TO_RT | DT_TO_RT | DATA_TO_RT | DATA_TO_DT_MZ_TO_RT_DT_RANGE | DATA_TO_DT_MZ_TO_RT_MZ_RANGE),
		ANY_TO_MZ = (RT_TO_MZ | DT_TO_MZ | DATA_TO_DT_MZ_DT_RANGE | DATA_TO_DT_MZ_MZ_RANGE | DATA_TO_DT_MZ_TO_MZ_DT_RANGE | DATA_TO_DT_MZ_TO_MZ_MZ_RANGE),
		ANY_TO_DT = (RT_TO_DT | MZ_TO_DT | DATA_TO_DT_MZ_DT_RANGE | DATA_TO_DT_MZ_MZ_RANGE | DATA_TO_DT_MZ_TO_DT_DT_RANGE | DATA_TO_DT_MZ_TO_DT_MZ_RANGE),
		ANY_TO_TIC_INT = (RT_TO_TIC_INT | MZ_TO_TIC_INT | DT_TO_TIC_INT | DATA_TO_DT_MZ_TO_TIC_INT),
	};


	struct IntegrationRange
	{
		double start;
		double end;
	};

	struct IntegrationTypeDesc
	{
		QString detailed;
		QString brief;
	};


	class HistoryItem
	{
		public:
			QDateTime m_dateTime;

			QMap<IntegrationType, IntegrationTypeDesc *> m_intTypeDescMap;

			QMap<IntegrationType, IntegrationRange *> m_integrationMap;

			HistoryItem();
 HistoryItem(const HistoryItem &other);
 HistoryItem &operator=(const HistoryItem &other);
 ~HistoryItem();

 void initializeIntegrationTypes();

 bool operator==(const HistoryItem &other) const;
 bool contains(IntegrationType intType,
 const IntegrationRange &intRange) const;

 void setDateTime();
 QDateTime dateTime() const;

 bool hasIntegrationType(IntegrationType intType);

 void newIntegrationRange(IntegrationType intType, double start, double end);
 IntegrationRange *integrationRange(IntegrationType intType) const;

 bool isValid() const;

 QString asText(bool brief = false) const;
 };


 class History
 {
 public:
 QList<HistoryItem *> m_itemList;

 History();
 History(const History &other);
 History &operator=(const History &other);
 ~History();

 void freeList();

 void appendHistory(const History &history, bool removeDuplicates = true);
 int appendHistoryItem(HistoryItem *item);
 int duplicateHistoryItem(const HistoryItem &item);

 HistoryItem *
 firstHistoryItem(IntegrationType integrationType,
 IntegrationRange **integrationRange = nullptr);
 HistoryItem *lastHistoryItem(IntegrationType integrationType,
 IntegrationRange **integrationRange = nullptr);
 int containsHistoryItem(const HistoryItem &item);


 bool innermostRtRange(double *start, double *end) const;
 bool innermostMzRange(double *start, double *end) const;
 bool innermostDtRange(double *start, double *end) const;
 bool innermostRange(IntegrationType integrationType, double *start,
 double *end) const;
 QString innermostRangesAsText() const;

 bool isValid() const;

 QString asText(const QString &header = QString(),
 const QString &footer = QString(), bool brief = false) const;
 };


} // namespace msXpertSuite

#endif /* HISTORY_HPP */
