/*   msXpertSuite - mass spectrometry software suite
 *   -----------------------------------------------
 *   Copyright(C) 2016 Filippo Rusconi
 *
 *   http://www.{site}.org
 *
 *   This file is part of the msXpertSuite project.
 *
 *   The msXpertSuite project is the successor of the massXpert project. This
 *   project now includes various independent modules:
 *   - massXpert, the original mass spectrometry environment software (greatly
 *     refactored code);
 *   - mineXpert, a powerful TIC chromatogram/mass/drift spectrum viewer/miner;
 *
 *    This program is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */


#ifndef MASS_SPEC_DATA_FILE_LOADER_MZML_HPP
#define MASS_SPEC_DATA_FILE_LOADER_MZML_HPP

#include <QtSql>
#include <QSqlDatabase>
#include <QStringList>
#include <QXmlStreamReader>
#include <QString>


#include "MassSpecDataSet.hpp"
#include "MassSpecDataFileLoader.hpp"
#include "MassSpecSqlite3Handler.hpp"
#include "MassSpecFileMetaData.hpp"

#include "History.hpp"

// For the compression enum and for the LogType enum.
#include "globals.hpp"


namespace msXpertSuite
{

  class MassSpecDataFileLoaderMzml : public MassSpecDataFileLoader
  {

    public:

    // We need this instance for all the SQLite3 database access (read/write)
    // work.

    MassSpecSqlite3Handler m_sqlite3Handler;
    MassSpecFileMetaData m_msFileMetaData;

    enum DataKind
    {
      DATA_KIND_UNSET,
      DATA_KIND_MZ,
      DATA_KIND_TIME,
      DATA_KIND_I,
    };

    DataKind        m_dataKind        = DataKind::DATA_KIND_UNSET;
    DataCompression m_dataCompression = DataCompression::DATA_COMPRESSION_NONE;


    IntegrationType      m_integrationType;
    History              m_history;

    // When loading data without passing a MassSpecDataSet *, then this is the
    // way to know that the loading should be in streamed mode. Otherwise, the
    // m_isStreamed member of MassSpecDataSet provides the information.

    bool m_isStreamed = false;

    bool m_isLoadToVectors = false;

    // We do not ownn the MassSpecDataSet instance !!!

    MassSpecDataSet *mp_massSpecDataSet = nullptr;

    // Helper map to store key/val pairs while doing the computations.

    QMap<double, double> m_keyValMap;

    // Helper mass spectrum pointer to store the address of the spectrum to be
    // allocated in a lasting manner so that it can be accessed across mzML
    // file-parsing functions. Only delete it locally, when its use it known to
    // be terminated.

		// Attention ! We may not own that new spectrum if it has been appended to
		// the mp_massSpecDataSet's list of spectra. We delete that spectrum only
		// when we know that we can do it, because it is no longer of use. This is
		// why we do not free that spectrum in the destructor.
    MassSpectrum *mp_newSpectrum = nullptr;

    // Helper mass spectrum to store the mass spectrum being created by
    // successive combinations.

    MassSpectrum m_massSpectrum;

    // Private functions that put the spectral data in various destinations.
    // These functions are called from within the MzML-parsing functions
    // (specifically the parse_spectrumList_element() function.
    int loadSpectrumToMassSpecDataSet();
    int loadSpectrumToVectors();

    int integrateToRt();
    int integrateToMz();
    int integrateToDt();


    QString formatDescription() const override;

    // Data loading
    static bool canLoadData(QString fileName);

    QXmlStreamReader m_xmlStreamReader;

    QStringList m_errorMessageList;

    // Pseudo-static stuff

    // Number of spectra actually stored in the database.
    int m_storedSpectraCount      = 0;
    int m_storedChromatogramCount = 0;

    // How many spectra do we parse before providing a feedback message on the
    // progression of the parsing of the file?
    int m_parseProgressionNumber = 100;

    // Number of parsing cycles. This value is reset each time the
    // m_parseProgNumb value is reached.
    int m_parseProgression = 0;

    // Number of spectra that were skipped, mainly because <binary></binary> was
    // empty.
    int m_totalSkippedSpectra = 0;

    // Set to true if one spectrum was encountered with an id mentioning that it
    // was from Function 2 of the Waters raw directory-file.
    bool m_wasFunction2Spectrum = false;

    // Spectral stuff

    // This is the index of the spectrum as read from the <spectrum> element.
    // This value is part of the record that is stored in the spectralData
    // database.
    int     m_spectrumIndex = -1;
    int     m_msLevel       = -1;
    QString m_spectrumTitle = "NOT_SET";

    // This is the value that describes in the MassSpecFileMetaData the actual
    // number of
    // stored spectra.
    int    m_specPeakCount                    = -1;
    int    m_scanCount                        = -1;
    double m_tic                              = -1;
    double m_scanStartTime                    = -1;
    int    m_spectrumBinaryDataArrayListCount = -1;

    // Mobility stuff
    int    m_driftDataId          = -1;
    bool   m_isMobilityExperiment = false;
    double m_curDriftTime         = -1;
    double m_driftTimeMin         = -1;
    double m_driftTimeMax         = -1;
    double m_driftTimeStep        = -1;
    double m_prevDriftTime        = -1;
    int    m_curDriftBin          = 0;
    int    m_driftBinCount        = -1;

    // Precursor ions stuff
    int           m_precursorIonId = -1;
    QList<double> m_precursorMzList;
    int           m_precursorListCount   = -1;
    double        m_precursorMz          = -1;
    int           m_selectedIonListCount = -1;
    double        m_selectedIonMz        = -1;

    // Scan stuff
    int        m_scanWindowListCount  = -1;
    double     m_scanWindowLowerLimit = -1;
    double     m_scanWindowUpperLimit = -1;
    double     m_collisionEnergy      = -1;
    double     m_encodedLength        = -1;
    QByteArray m_mzQByteArray;
    QByteArray m_iQByteArray;

    // Chromatogram stuff
    QString    m_chromatogramTitle;
    int        m_chromatogramListCount                = -1;
    int        m_chromatogramIndex                    = -1;
    int        m_chromatogramBinaryDataArrayListCount = -1;
    QByteArray m_chromTimeQByteArray;
    QByteArray m_chromIQByteArray;

    // The entry point to the mzMl file:
    void parseMzmlFile();

    // All the xml document elements parsing functions.
    void         parse_indexedmzML_element();
    void         parse_mzML_element();
    virtual void parse_run_element();
    void         parse_spectrumList_element();
    void         parse_spectrum_element();
    void         parse_spectrum_cvParam_element();
    void         parse_scanList_element();
    void         parse_scanList_cvParam_element();
    void         parse_scanList_scan_element();
    void         parse_scan_cvParam_element();
    void         parse_scanWindowList_element();
    void         parse_scanWindow_element();
    void         parse_precursorList_element();
    void         parse_precursor_element();
    void         parse_isolationWindow_element();
    void         parse_selectedIonList_element();
    void         parse_selectedIon_element();
    void         parse_activation_element();
    void         parse_spectrum_binaryDataArrayList_element();
    void         parse_spectrum_binaryDataArray_element();
    void         parse_spectrum_binaryDataArray_cvParam_element();
    void         parse_chromatogramList_element();
    void         parse_chromatogram_element();
    void         parse_chromatogram_cvParam_element();
    void         parse_chromatogram_binaryDataArrayList_element();
    void         parse_chromatogram_binaryDataArray_cvParam_element();
    void         parse_chromatogram_binaryDataArray_element();


    void resetSpectralData();
    void resetMobilityData();
    void resetChromatographicData();

    void seedMobilityData();

    MassSpecDataFileLoaderMzml(const QString &fileName = QString());

    virtual ~MassSpecDataFileLoaderMzml();

    void setFileName(const QString &fileName);
    QString fileName() const;

    void setMassSpecDataSet(MassSpecDataSet *massSpecDataSet);

    void setParseProgressionNumber(int value);
    bool containsMobilityData();
    int  readSpectrumCount() override;
    int  spectrumCountNoFunction2();

    int loadData(MassSpecDataSet *massSpecDataSet);
    int loadData(const QString &dbFileName = QString());
    int loadData(const History &history, IntegrationType integrationType,
                 QVector<double> *keyVector, QVector<double> *valVector);

    int streamedIntegration(const History &  history,
                            IntegrationType  integrationType,
                            QVector<double> *keyVector,
                            QVector<double> *valVector);
  };

} // namespace msXpertSuite

#endif /* MASS_SPEC_DATA_FILE_LOADER_MZML_HPP */
