/*   msXpertSuite - mass spectrometry software suite
 *   -----------------------------------------------
 *   Copyright(C) 2016 Filippo Rusconi
 *
 *   http://www.{site}.org
 *
 *   This file is part of the msXpertSuite project.
 *
 *   The msXpertSuite project is the successor of the massXpert project. This
 *   project now includes various independent modules:
 *   - massXpert, the original mass spectrometry environment software (greatly
 *     refactored code);
 *   - mobXpert, a module to process and mine mobility mass spectrometry data;
 *   - viewXpert, a powerful TIC chromatogram/mass spectrum viewer/miner;
 *
 *    This program is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */


#include <omp.h>
#include <limits>


/////////////////////// Qt includes
#include <QApplication> // for the clipboard
#include <QClipboard>
#include <QDebug>
#include <QFile>
#include <qmath.h>


/////////////////////// Local includes
#include "MassSpectrum.hpp"
#include "MassSpectrumList.hpp"
#include "Base64EncDec.hpp"


namespace msXpertSuite
{

	MassSpectrum::MassSpectrum(const QString &title)
	{
		m_title = title;
	}


	bool
		MassSpectrum::setupBinnability(const MassSpectrumList &massSpectra, bool reset)
		{
			if(massSpectra.size() == 0)
				return false;

			m_isBinnable = massSpectra.sameLengthSpectra();

			//qDebug() << __FILE__ << __LINE__
				//<< "Spectrum has same-length spectra:" << m_isBinnable
				//<< "and might be binnable:" << m_isBinnable;

			// We only do setup the bins, if there is no single spectrum in *this
			// spectrum or we are asked to reset it.

			if(m_isBinnable && (!size() || reset))
			{
				double minMz = 0;
				double maxMz = 0;

				QList<double> binList;

				massSpectra.characters(&minMz, &maxMz, nullptr, nullptr, nullptr);

				m_binMinMz = minMz;
				m_binMaxMz = maxMz;

				// Use the first spectrum of the list to fill in the binList

				if(qIsNaN(massSpectra.first()->fillInBins(&binList)))
					return false;

				//qDebug() << __FILE__ << __LINE__
					//<< QString("First spectrum to combine has %1 peaks and %2 bins.")
					//.arg(massSpectra.first()->size())
					//.arg(binList.size())
					//<< binList;

				m_isBinnable = isBinnable(binList);

				//qDebug() << __FILE__ << __LINE__
					//<< "Spectrum is binnable:" << m_isBinnable;

				if(m_isBinnable)
				{

					// Now that we know the characteristics of the mass spec list, we can
					// configure *this spectrum if not done already, that is, if size() is 0.

					m_binCount = binList.size();

					//qDebug() << __FILE__ << __LINE__
						//<< "minMz:" << minMz
						//<< "maxMz:" << maxMz
						//<< "spectrum size:" << size()
						//<< "m_binCount:" << m_binCount;

					setupBins(binList);
				}
			}

			return true;
		}


	double 
		MassSpectrum::fillInBins(QList<double> *binList)
		{
			if(binList == nullptr)
				qFatal("Fatal error at %s@%d. Program aborted.",
						__FILE__, __LINE__);

			// We need at least two mass peaks in the spectrum to compute the step
			// between the second and the first.

			if(size() < 2)
			{
				return qSNaN();
			}

			double greatestStep = 0;
			double tempStep = 0;

			double prevMz = at(0)->m_mz;

			for(int iter = 1; iter < size(); ++iter)
			{
				double curMz = at(iter)->m_mz;

				tempStep = curMz - prevMz;

				// Spectra are ascending-ordered for the mz value. If not, that's fatal.

				if(tempStep < 0)
					qFatal("Fatal error at %s@%d. Program aborted.",
							__FILE__, __LINE__);

				binList->append(tempStep);

				//QString msg = QString("Append step: %1 from computation %2 - %3 with result: %4")
				//.arg(tempStep, 0, 'f', 10)
				//.arg(curMz, 0, 'f', 10)
				//.arg(prevMz, 0, 'f', 10)
				//.arg(curMz - prevMz, 0, 'f', 10);

				//qDebug() << __FILE__ << __LINE__ << msg;

				if(tempStep > greatestStep)
					greatestStep = tempStep;

				prevMz = curMz;
			}

			//qDebug() << __FILE__ << __LINE__
			//<< "Number of elements in binList:" << binList->size();

#if 0

			qDebug() << __FILE__ << __LINE__
				<< "Writing the list of bins in file /tmp/initial-bins.txt";

			QFile file("/tmp/initial-bins.txt");
			file.open(QIODevice::WriteOnly);

			QTextStream fileStream(&file);

			for(int iter = 0; iter < binList->size(); ++iter)
			{

				fileStream << QString("%1 %2\n")
					.arg(binList->at(iter), 0, 'f', 10)
					.arg(0);
			}

			fileStream.flush();
			file.close();

#endif



			// Only set to the member datum if that was not computed already.  To set
			// it forcefully, use the other function.

			if(qIsNaN(m_greatestStep))
				m_greatestStep = greatestStep;

			return greatestStep;
		}


	double
		MassSpectrum::setGreatestStep(QList<double> *binList)
		{

			// Recompute the greatest step and forcefully set it to the member datum.

			m_greatestStep = fillInBins(binList);

			return m_greatestStep;
		}


	bool 
		MassSpectrum::isBinnable(const QList<double> &binList)
		{

			// A spectrum is binnable if when looking at all the m/z intervals in it,
			// they are all relatively similar and much less than unity (<< 1).

			if(binList.isEmpty())
				return false;

			int binCount = binList.size();

			double binSum = 0;
			double binAvg = 0;

			for(int iter = 0; iter < binCount; ++iter)
			{
				binSum += binList.at(iter);
			}

			binAvg = binSum / binCount;

			// If a spectrum has not a proper m/z value ladder it will not be
			// binnable. This is typically the case with the spectra obtained from
			// Water's Synapt MS in mobility mode. Below is a rough test.

			if(binAvg > 0.1)
				return false;

			double binVarN = 0;

			for(int iter = 0; iter < binCount; ++iter)
			{
				binVarN += (binList.at(iter) - binAvg) * (binList.at(iter) - binAvg);
			}

			double binVar = binVarN / binCount;

			//double binStdDev = sqrt(binVar);

			//QString msg = QString("binCount = %1 - binSum: %2 - binAvg = %3 - binVarN = %4"
			//" - binVar = %5 - binStdDev = %6\n")
			//.arg(binCount)
			//.arg(binSum, 0, 'f', 10)
			//.arg(binAvg, 0, 'f', 10)
			//.arg(binVarN, 0, 'f', 10)
			//.arg(binVar, 0, 'f', 10)
			//.arg(binStdDev, 0, 'f', 10);

			//qDebug() << __FILE__ << __LINE__
			//<< msg;

			if(binVar < 0.001)
				return true;

			return false;
		}


	double 
		MassSpectrum::determineMzShift(const MassSpectrum &massSpectrum)
		{
			if(!size() || !massSpectrum.size())
				return qSNaN();

			m_mzShift = massSpectrum.first()->m_mz - first()->m_mz;

			//qDebug() << __FILE__ << __LINE__
			//<< "m_mzShift:" << m_mzShift;

			return m_mzShift;
		}


	void
		MassSpectrum::setupBins(const QList<double> &binList)
		{
			if(size())
				qFatal("Fatal error at %s@%d. Program aborted.",
						__FILE__, __LINE__);

			if(binList.size() != m_binCount)
			{
				QString msg = QString("binList size: %1 -- m_binCount: %2\n")
					.arg(binList.size()).arg(m_binCount);

				qFatal("Fatal error at %s@%d with msg: %s. Program aborted.",
						__FILE__, __LINE__, msg.toLatin1().data());
			}

			// Seed the mass spectrum with the very first mz value that is in the bin
			// list. Thus, the for loop then starts at index 1.

			append(new MassPeak(m_binMinMz, 0));
			double previousMzBin = m_binMinMz;

			for(int iter = 0; iter < m_binCount; ++iter)
			{
				double tempMzBin = previousMzBin + binList.at(iter);

				//QString msg = QString("Append bin mz: %1 from computation %2 + %3 with result: %4")
				//.arg(tempMzBin, 0, 'f', 10)
				//.arg(previousMzBin, 0, 'f', 10)
				//.arg(binList.at(iter), 0, 'f', 10)
				//.arg(previousMzBin + binList.at(iter), 0, 'f', 10);

				//qDebug() << __FILE__ << __LINE__ << msg;

				previousMzBin = tempMzBin;

				append(new MassPeak(previousMzBin, 0));
			}

#if 0

			qDebug() << __FILE__ << __LINE__
				<< "Writing the list of bins in file /tmp/bins.txt";

			QFile file("/tmp/bins.txt");
			file.open(QIODevice::WriteOnly);

			QTextStream fileStream(&file);

			for(int iter = 0; iter < m_binCount; ++iter)
			{

				fileStream << QString("%1 %2\n")
					.arg(at(iter)->m_mz, 0, 'f', 10)
					.arg(0);
			}

			fileStream.flush();
			file.close();

#endif

			//qDebug() << __FILE__ << __LINE__
			//<< "Prepared bins with " << size() << "elements."
			//<< "starting with mz" << first()->m_mz
			//<< "ending with mz" << last()->m_mz;
		}


	int 
		MassSpectrum::whichBinIndex(double mz)
		{

			//qDebug() << __FILE__ << __LINE__
			//<< "Entering whichBinIndex with mz:" 
			//<< QString("%1").arg(mz, 0, 'f', 10)
			//<< "and m_lastBinIdx:" << m_lastBinIdx;

			int lastTestIndex = 0;

			int returnIndex = -1;

			if(!size())
				return 0;

			for(int iter = m_lastBinIdx; iter < size(); ++iter)
			{
				double iterMz = at(iter)->m_mz ;

				//qDebug() << __FILE__ << __LINE__
				//<< "whichBinIndex for mz:" << QString("%1").arg(mz, 0, 'f', 10)
				//<< "spectrum iter:" << iter 
				//<< "with mz:" << QString("%1").arg(iterMz, 0, 'f', 10);

				if(mz >= iterMz)
				{
					if(mz == iterMz)
					{
						returnIndex = iter;
						break;
					}
					else
					{
						// If this is the last bin, then we need to return iter.

						if(iter == size())
						{
							returnIndex = iter;
							break;
						}
						else
							lastTestIndex = iter;
					}
				}
				else if(mz < iterMz)
				{
					returnIndex = lastTestIndex;
					break;
				}
				else
					qFatal("Fatal error at %s@%d. Program aborted.",
							__FILE__, __LINE__);
			}

			// Due to binning, we might be here if mz was greater than the last mz of
			// the spectrum. In this case we just return the last spectrum index.

			if(returnIndex == -1)
			{
				returnIndex = size() - 1;

				//qDebug() << __FILE__ << __LINE__
				//<< "needed to set returnIndex to size() - 1;"
				//<< "mz: " << QString("%1").arg(mz, 0, 'f', 10)
				//<< "bin index:" << returnIndex
				//<< "with bin's mz:" << QString("%1").arg(at(returnIndex)->m_mz, 0, 'f', 10)
				//<< "bin count: size():" << size()
				//<< "last bin's mz:" << QString("%1").arg(at(size() - 1)->m_mz, 0, 'f', 10);
			}

			m_lastBinIdx = returnIndex;

			//qDebug() << __FILE__ << __LINE__
			//<< "mz: " << QString("%1").arg(mz, 0, 'f', 10)
			//<< "bin index:" << returnIndex
			//<< "with bin's mz:" << QString("%1").arg(at(returnIndex)->m_mz, 0, 'f', 10)
			//<< "bin count: size():" << size()
			//<< "last bin's mz:" << QString("%1").arg(at(size() - 1)->m_mz, 0, 'f', 10);


			return returnIndex;
		}


	void
		MassSpectrum::initialize(const QList<double> &mzList,
				const QList<double> &iList)
		{
			if(mzList.size() != iList.size())
				qFatal("Fatal error at %s@%d. Program aborted.", __FILE__, __LINE__);

			deleteMassPeaks();

			for(int iter = 0; iter < mzList.size(); ++iter)
			{
				MassPeak *massPeak = new MassPeak(mzList.at(iter), iList.at(iter));
				append(massPeak);
			}
		}


	void
		MassSpectrum::initialize(const QList<double> &mzList,
				const QList<double> &iList, double mzStart,
				double mzEnd)
		{
			if(mzList.size() != iList.size())
				qFatal("Fatal error at %s@%d. Program aborted.", __FILE__, __LINE__);

			deleteMassPeaks();

			for(int iter = 0; iter < mzList.size(); ++iter)
			{
				double mz = mzList.at(iter);

				if(mz >= mzStart && mz <= mzEnd)
				{
					MassPeak *massPeak = new MassPeak(mz, iList.at(iter));
					append(massPeak);
				}
			}
		}


	void
		MassSpectrum::initialize(const QList<double> &mzList,
				const QList<double> &iList, double driftTime)
		{
			if(mzList.size() != iList.size())
				qFatal("Fatal error at %s@%d. Program aborted.", __FILE__, __LINE__);

			deleteMassPeaks();

			for(int iter = 0; iter < mzList.size(); ++iter)
			{
				MassPeak *massPeak = new MassPeak(mzList.at(iter), iList.at(iter));
				append(massPeak);
			}

			m_dt = driftTime;
		}


	void
		MassSpectrum::initialize(const QList<double> &mzList,
				const QList<double> &iList, double driftTime,
				double mzStart, double mzEnd)
		{
			if(mzList.size() != iList.size())
				qFatal("Fatal error at %s@%d. Program aborted.", __FILE__, __LINE__);

			deleteMassPeaks();

			for(int iter = 0; iter < mzList.size(); ++iter)
			{
				double mz = mzList.at(iter);

				if(mz >= mzStart && mz <= mzEnd)
				{
					MassPeak *massPeak = new MassPeak(mz, iList.at(iter));
					append(massPeak);
				}
			}

			m_dt = driftTime;
		}


	void 
		MassSpectrum::initialize(const QVector<double> &mzVector, const QVector<double> &iVector)
		{
			if(mzVector.size() != iVector.size())
				qFatal("Fatal error at %s@%d. Program aborted.", __FILE__, __LINE__);

			deleteMassPeaks();

			for(int iter = 0; iter < mzVector.size(); ++iter)
				append(new MassPeak(mzVector.at(iter), iVector.at(iter)));
		}


	MassSpectrum::~MassSpectrum()
	{
		deleteMassPeaks();
	}


	void
		MassSpectrum::reset()
		{
			deleteMassPeaks();
			m_title.clear();

			m_lastModifIdx = 0;
			m_dt           = 0;
			m_rt           = 0;
		}


	void
		MassSpectrum::deleteMassPeaks()
		{
#pragma omp parallel for
			// This for loop has the advantage over the while(!isEmpty){delete
			// takeFirst()} variant that it accessed the QList<MassPeak *> items in a
			// parallel manner. But we then need to actually empty the base class of all
			// the delete item pointers.
			for(int iter = 0; iter < size(); ++iter)
			{
				// qDebug() << __FILE__ << __LINE__
				// << "Deleting MassPeak at " << iter;

				delete at(iter);
			}

			// Call the parent class clear() to actually empty the QList<MassPeak>
			clear();
		}


	double
		MassSpectrum::tic()
		{
			// Iterate in each mass peak of this spectrum and sum up its intensity with
			// all the other peaks' intensity. Return that value.

			double tic = 0;

#pragma omp parallel
			{
				// int maxNumbThreads = omp_get_max_threads();
				// omp_set_num_threads(10);
				// int numbThreads = omp_get_num_threads();
				// qDebug()   << __FILE__ << __LINE__
				// << "Number of threads: " << numbThreads << "over" << maxNumbThreads;

#pragma omp for reduction(+ : tic)
				// 20160921 - This code was tested to provide identical results to the
				// serial version
				// and to provide an increase in the execution speed of roughly 2x with
				// 4 threads.

				for(int iter = 0; iter < size(); ++iter)
				{
					// int numbThreads = omp_get_num_threads();
					// qDebug()   << __FILE__ << __LINE__
					// << "Number of threads: " << numbThreads;

					tic += at(iter)->m_i;
				}
			}

			return tic;
		}


	double
		MassSpectrum::tic(double mzStart, double mzEnd)
		{
			// Iterate in each mass peak of this spectrum and sum up its intensity with
			// all the other peaks' intensity. Return that value. Do that only if the
			// mz value is in the ranged passed as parameters.

			double tic = 0;

			for(int iter = 0; iter < size(); ++iter)
			{
				double mz = at(iter)->m_mz;

				if(mz >= mzStart && mz <= mzEnd)
					tic += at(iter)->m_i;
			}

			return tic;
		}



	int
		MassSpectrum::combine(const MassPeak &massPeak)
		{
			//qDebug() << __FILE__ << __LINE__
			//<< "Entering combine";

			// This function only provide very fast additions, insertion, because a
			// mass spectrum is inherently sorted in increasing order. Otherwise this
			// function would not work.

			// Returns 0 if the peak was not added but only the intensity was
			// increment, and 1 if the peak was actually added (either prepended,
			// inserted or appended).

			// We get a MassPeak instance and we must integrate it into *this list. If
			// a mass peak by the same mz is found, then the intensity is increased.
			// Otherwise the mass peak is integrated in such a manner that the
			// spectrum sorts correctly in ascending order with respect to mz values.

			// The index of the modified element of the list, be that modification an
			// insertion or an intensity increment, the m_lastModifIdx is set to that
			// index such that the next call can make use of it to accelerate the
			// search process.

			// Take into account the m_mzShift that might have computed a non-0 value
			// to align the new spectrum to *this one.

			double mz = massPeak.m_mz + m_mzShift;

			// Check if decimals are specified.
			if(m_requiredDecimals != -1)
			{
				mz = ceil((mz * pow(10, m_requiredDecimals)) - 0.49) / pow(10, m_requiredDecimals);

				// if no decimal at all is required: int mz = (double) (int) mz;

				// qDebug("Rounding from %.5f to %.5f\n",
				// massPeak.m_mz, mz);
			}

			if(!size())
			{
				append(new MassPeak(mz, massPeak.m_i));

				m_lastModifIdx = 0;
				return 1;
			}

			if(size() == 1)
			{
				if(mz == last()->m_mz)
				{
					last()->m_i += massPeak.m_i;

					m_lastModifIdx = 0;
					return 0;
				}
				else if(mz > last()->m_mz)
				{
					append(new MassPeak(mz, massPeak.m_i));

					m_lastModifIdx = 1;
					return 1;
				}
				else
				{
					prepend(new MassPeak(mz, massPeak.m_i));

					m_lastModifIdx = 1;
					return 1;
				}
			}

			// At this point we know that there are at least two MassPeak in this
			// spectrum.

			if(m_lastModifIdx >= size())
			{
				m_lastModifIdx = size() / 2;
			}

			if(mz < at(m_lastModifIdx)->m_mz)
			{
				m_lastModifIdx = 0;
			}

			int oneLessThanSpecSize = size() - 1;

			for(int iter = m_lastModifIdx; iter < oneLessThanSpecSize; ++iter)
			{
				if(mz < at(iter)->m_mz)
				{
					insert(iter, new MassPeak(mz, massPeak.m_i));
					m_lastModifIdx = iter;
					return 1;
				}

				if(mz == at(iter)->m_mz)
				{

					// There is a MassPeak that has that same mz, only simply increment
					// the
					// intensity.

					at(iter)->m_i += massPeak.m_i;
					m_lastModifIdx = iter;
					return 0;
				}

				if(mz > at(iter)->m_mz && mz < at(iter + 1)->m_mz)
				{
					insert(iter + 1, new MassPeak(mz, massPeak.m_i));
					m_lastModifIdx = iter + 1;
					return 1;
				}
			}

			// At this point, we did not insert the mass peak, nor did we increment a
			// preexisting mass peak. There are two possibilities:

			if(mz == last()->m_mz)
			{
				last()->m_i += massPeak.m_i;
				m_lastModifIdx = size() - 1;
				return 0;
			}
			else
			{
				// The mz value is greater than the last value of the
				// spectrum. Simply perform an append.
				append(new MassPeak(mz, massPeak.m_i));
				m_lastModifIdx = size() - 1;
				return 1;
			}
		}


	int
		MassSpectrum::combineBinned(const MassPeak &massPeak)
		{
			//qDebug() << __FILE__ << __LINE__
			//<< "Entering combineBinned";

			if(!size())
				qFatal("Fatal error at %s@%d. Program aborted.",
						__FILE__, __LINE__);

			double mz = massPeak.m_mz;

			if(m_requiredDecimals != -1)
			{
				mz = ceil((mz * pow(10, m_requiredDecimals)) - 0.49) / pow(10, m_requiredDecimals);

				// if no decimal at all is required: int mz = (double) (int) mz;

				qDebug("%s %d Rounding from %.5f to %.5f\n",
						__FILE__, __LINE__,
						massPeak.m_mz, mz);
			}

			int binIndex = whichBinIndex(mz);

			if(binIndex == -1)
				qFatal("Fatal error at %s@%d. Program aborted.",
						__FILE__, __LINE__);

			//QString msg = QString("must combine mass peak of mz: %1"
			//"with bin of mz,i: (%2, %3)")
			//.arg(mz, 0, 'f', 10)
			//.arg(at(binIndex)->m_mz, 0, 'f', 10)
			//.arg(at(binIndex)->m_i, 0, 'f', 10);

			//qDebug() << __FILE__ << __LINE__ << msg;

			MassPeak *targetMassPeak = at(binIndex);

			targetMassPeak->m_i += massPeak.m_i;

			//msg = QString("after combination: (mz,i): (%1,%2)")
			//.arg(at(binIndex)->m_mz, 0, 'f', 10)
			//.arg(at(binIndex)->m_i, 0, 'f', 10);

			//qDebug() << __FILE__ << __LINE__ << msg;

			return binIndex;
		}


	int
		MassSpectrum::combineSlow(const MassPeak &massPeak)
		{
			// Returns 0 if the peak was not added but only the intensity was increment,
			// and
			// 1 if the peak was actually added (either prepended inserted or appended).

			// We get a MassPeak instance and we must integrate it into *this list. If
			// a mass peak by the same mz is found, then the intensity is increased.
			// Otherwise the mass peak is integrated in such a manner that the
			// spectrum sorts correctly in ascending order with respect to mz values.

			double mz = massPeak.m_mz + m_mzShift;

			if(m_requiredDecimals != -1)
			{
				mz = ceil((mz * pow(10, m_requiredDecimals)) - 0.49) / pow(10, m_requiredDecimals);

				// if no decimal at all is required: int mz = (double) (int) mz;

				// qDebug("Rounding from %.5f to %.5f\n",
				// massPeak.m_mz, mz);
			}

			if(!size())
			{
				append(new MassPeak(mz, massPeak.m_i));
				return 1;
			}

			if(size() == 1)
			{
				if(mz == last()->m_mz)
				{
					last()->m_i += massPeak.m_i;
					return 0;
				}
				else if(mz > last()->m_mz)
				{
					append(new MassPeak(mz, massPeak.m_i));
					return 1;
				}
				else
				{
					prepend(new MassPeak(mz, massPeak.m_i));
					return 1;
				}
			}

			// Now that we have handled two simple cases, go on with the "sorting".

			int oneLessThanSpecSize = size() - 1;

			for(int iter = 0; iter < oneLessThanSpecSize; ++iter)
			{
				if(mz < at(iter)->m_mz)
				{
					insert(iter, new MassPeak(mz, massPeak.m_i));
					return 1;
				}

				if(mz == at(iter)->m_mz)
				{
					// There is a MassPeak that has that same mz, only simply increment
					// the
					// intensity.
					at(iter)->m_i += massPeak.m_i;
					return 0;
				}

				if(mz > at(iter)->m_mz && mz < at(iter + 1)->m_mz)
				{
					insert(iter + 1, new MassPeak(mz, massPeak.m_i));
					return 1;
				}
			}

			// At this point, we did not insert the mass peak, nor did we increment a
			// preexisting mass peak. There are two possibilities:

			if(mz == last()->m_mz)
			{
				last()->m_i += massPeak.m_i;
				return 0;
			}
			else
			{
				// The mz value is greater than the last value of the
				// spectrum. Simply perform an append.
				append(new MassPeak(mz, massPeak.m_i));
				return 1;
			}
		}


	int
		MassSpectrum::combine(const MassSpectrum &massSpectrum)
		{
			// We receive a spectrum and we need to combine this spectrum in this mass
			// spectrum.

			//qDebug() << __FILE__ << __LINE__
			//<< "Starting combination of the spectrum with "
			//<< massSpectrum.size() << "peaks in it.";

			if(m_isBinnable)
				determineMzShift(massSpectrum);

			// Reset the index anchors
			m_lastBinIdx = 0;
			m_lastModifIdx = 0;

			int iter = 0;

			for(iter = 0; iter < massSpectrum.size(); ++iter)
			{
				if(m_isBinnable)
					combineBinned(*massSpectrum.at(iter));
				else
					combine(*massSpectrum.at(iter));
			}

			// qDebug() << __FILE__ << __LINE__
			// << "Done combining a new spectrum in this spectrum, that has now "
			// << size() << "mass peaks";

			// Return the number of combined mass peaks.
			return iter;
		}


	int
		MassSpectrum::combine(const MassSpectrum &massSpectrum, 
				double mzStart, double mzEnd)
		{
			// We receive a spectrum and we need to combine this spectrum in this mass
			// spectrum. However, we need to check each individual MassPeak of the
			// MassSpectrum instance for their m/z value, because it can only be
			// combined if its value is in between both mzStart and mzEnd arguments to
			// this function.

			// qDebug() << __FILE__ << __LINE__
			// << "Starting combination of the spectrum with " <<
			// << massSpectrum.size()
			// << "peaks in it.";

			// Reset the index anchors
			m_lastBinIdx = 0;
			m_lastModifIdx = 0;

			int iter = 0;

			for(; iter < massSpectrum.size(); ++iter)
			{
				MassPeak *iterMassPeak = massSpectrum.at(iter);

				if(iterMassPeak->m_mz >= mzStart && iterMassPeak->m_mz <= mzEnd)
				{
					if(m_isBinnable)
						combineBinned(*iterMassPeak);
					else
						combine(*iterMassPeak);
				}
				// Because the spectra are ordered, if the currently iterated mass peak
				// has a mass greater than the accepted mzEnd, then break the loop as
				// this cannot be any better later.
				else if(iterMassPeak->m_mz > mzEnd)
					break;
			}
			// qDebug() << __FILE__ << __LINE__
			// << "Done combining a new spectrum in this spectrum, current with "
			// << size() << "mass peaks";

			// Return the number of combined mass peaks.
			return iter;
		}


#if 0

	// For the record, the function used before when we did not encounter
	// the Bruker microQTof spectra that were misaligned and which required
	// binning for proper combination.

	int
		MassSpectrum::combine(const MassSpectrumList &massSpectra)
		{
			// We receive a list of mass spectra and wee need to combine all the
			// spectra thereof in this mass spectrum.

			// qDebug() << __FILE__ << __LINE__
			// << "Starting combination of " << massSpectra.size()
			// << "mass spectra into this mass spectrum";

			int specCount = massSpectra.size();

			int combinedSpectra = 0;

			for(int iter = 0; iter < specCount; ++iter)
			{
				MassSpectrum *massSpectrum = massSpectra.at(iter);

				combine(*massSpectrum);

				combinedSpectra++;
			}

			// qDebug() << __FILE__ << __LINE__
			// << "Effectively combined " << combinedSpectra
			// << "in this single spectrum for a total of "
			// << size() << "mass peaks";

			return combinedSpectra;

		}
#endif


	int
		MassSpectrum::combine(const MassSpectrumList &massSpectra, bool reset /* = false */)
		{
			// We receive a list of mass spectra and wee need to combine all the
			// spectra thereof in this mass spectrum.

			if(!massSpectra.size())
				return 0;

			// qDebug() << __FILE__ << __LINE__
			// << "Starting combination of " << massSpectra.size()
			// << "mass spectra into this mass spectrum";

			// This function might be called recursively with various mass spectrum
			// lists. We only need to seed the bin combination data if:
			// 1. that has not been done already
			// 2. reset is true

			if(!size() || reset)
				if(!setupBinnability(massSpectra, reset))
					qFatal("Fatal error at %s@%d. Program aborted.",
							__FILE__, __LINE__);

			MassSpectrum *massSpectrum = nullptr;
			int specCount = massSpectra.size();

			int combinedSpectra = 0;

			for(int iter = 0; iter < specCount; ++iter)
			{
				massSpectrum = massSpectra.at(iter);

				combine(*massSpectrum);

#if 0
				QString fileName = QString("/tmp/combined-spectrum-after-%1-combinations.txt")
					.arg(iter + 1);

				QFile file(fileName);

				file.open(QIODevice::WriteOnly);

				QTextStream fileStream(&file);

				for(int iter = 0; iter < size(); ++iter)
				{
					fileStream << QString("%1 %2\n")
						.arg(at(iter)->m_mz, 0, 'f', 10)
						.arg(at(iter)->m_i, 0, 'f', 10);
				}

				fileStream.flush();
				file.close();
#endif


				combinedSpectra++;
			}
			// qDebug() << __FILE__ << __LINE__
			// << "Effectively combined " << combinedSpectra
			// << "in this single spectrum for a total of "
			// << size() << "mass peaks";

			return combinedSpectra;
		}


	int
		MassSpectrum::combine(const MassSpectrumList &massSpectra,
				double mzStart, double mzEnd, bool reset)
		{
			// We receive a list of mass spectra, that is a list of (list of
			// MassPeak*) and wee need to combine all the spectra thereof in this mass
			// spectrum. However, we need to check each individual MassPeak of the
			// MassSpectrum instances for their m/z value, because it can only be
			// combined if its value is in between both mzStart and mzEnd arguments to
			// this function.

			if(massSpectra.size() == 0)
				return 0;

			//qDebug() << __FILE__ << __LINE__
				//<< "Starting combination of " << massSpectra.size()
				//<< "mass spectra into this mass spectrum";

			if(!size() || reset)
				if(!setupBinnability(massSpectra, reset))
					qFatal("Fatal error at %s@%d. Program aborted.",
							__FILE__, __LINE__);

			int combinedSpectra = 0;

			for(int iter = 0; iter < massSpectra.size(); ++iter)
			{
				MassSpectrum *iterMassSpectrum = massSpectra.at(iter);

				// qDebug() << __FILE__ << __LINE__
				// << "Currently processed mass spectrum has"
				// << iterMassSpectrum->size() << "peaks";

				for(int jter = 0; jter < iterMassSpectrum->size(); ++jter)
				{
					MassPeak *iterMassPeak = iterMassSpectrum->at(jter);

					if(iterMassPeak->m_mz >= mzStart && iterMassPeak->m_mz <= mzEnd)
						combine(*iterMassPeak);
				}

				// qDebug() << __FILE__ << __LINE__
				// << "Done combining a new spectrum in this spectrum, current with "
				// << size() << "mass peaks";

				combinedSpectra++;
			}

			// qDebug() << __FILE__ << __LINE__
			// << "Effectively combined " << combinedSpectra
			// << "in this single spectrum for a total of "
			// << size() << "mass peaks";

			return combinedSpectra;
		}


	int
		MassSpectrum::combine(const QList<double> &mzList, const QList<double> &iList)
		{
			// We get two lists of mz and i, typically out of the data file, and we need
			// to
			// construct (or continue constructing) a mass spectrum. For each mz in the
			// mzList, we iterate in the spectrum and we check if that mz exists. If so
			// the
			// intensity of the found MassPeak is increased. If not a new MassPeak is
			// inserted at the proper place with the corresonding intensity from iList.

			// Sanity check

			int listSize = mzList.size();
			if(listSize != iList.size())
				qFatal("Fatal error at %s@%d. Program aborted.", __FILE__, __LINE__);

			int count = 0;

			for(int iter = 0; iter < listSize; ++iter)
			{

				count += combine(MassPeak(mzList.at(iter), iList.at(iter)));
			}

			return count;
		}


	int
		MassSpectrum::combine(const QList<double> &mzList, const QList<double> &iList,
				double mzStart, double mzEnd)
		{
			// Sanity check

			int listSize = mzList.size();
			if(listSize != iList.size())
				qFatal("Fatal error at %s@%d. Program aborted.", __FILE__, __LINE__);

			int count = 0;

			for(int iter = 0; iter < listSize; ++iter)
			{
				double mz = mzList.at(iter);

				if(mz >= mzStart && mz <= mzEnd)
				{
					count += combine(MassPeak(mzList.at(iter), iList.at(iter)));
				}
			}

			return count;
		}


	int
		MassSpectrum::combineIndices(const MassSpectrumList &massSpectra,
				int firstIdx, int lastIdx)
		{
			// We receive a list of mass spectra and wee need to combine all the
			// spectra thereof in this mass spectrum.

			// qDebug() << __FILE__ << __LINE__
			// << "Starting combination of " << massSpectra.size()
			// << "mass spectra into this mass spectrum";

			int specCount = massSpectra.size();

			// Remove one specific case:
			if(!specCount)
				return 0;

			// Check sanity of the parameters.
			if(firstIdx < -1 || lastIdx < -1 || firstIdx > specCount - 1 ||
					lastIdx > specCount - 1)
				qFatal("Fatal error at %s@%d. Program aborted.", __FILE__, __LINE__);

			int iterFirst = firstIdx;
			int iterLast  = lastIdx;

			if(firstIdx == -1)
				iterFirst = 0;
			if(lastIdx == -1)
				iterLast = specCount;

			int combinedSpectra = 0;

			for(int iter = iterFirst; iter < iterLast; ++iter)
			{
				MassSpectrum *massSpectrum = massSpectra.at(iter);

				combine(*massSpectrum);

				combinedSpectra++;
			}

			// qDebug() << __FILE__ << __LINE__
			// << "Effectively combined " << combinedSpectra
			// << "in this single spectrum for a total of "
			// << size() << "mass peaks";

			return combinedSpectra;
		}


	int
		MassSpectrum::contains(double mz) const
		{
			int specSize = size();

			for(int iter = 0; iter < specSize; ++iter)
				if(at(iter)->m_mz == mz)
					return iter;

			return -1;
		}


	QList<double>
		MassSpectrum::mzList() const
		{
			QList<double> list;

			for(int iter = 0; iter < size(); ++iter)
			{
				list.append(at(iter)->m_mz);
			}

			return list;
		}


	QList<double>
		MassSpectrum::iList() const
		{
			QList<double> list;

			for(int iter = 0; iter < size(); ++iter)
			{
				list.append(at(iter)->m_i);
			}

			return list;
		}


	void
		MassSpectrum::lists(QList<double> *mzList, QList<double> *iList) const
		{
			if(!mzList)
				qFatal("Fatal error at %s@%d. Program aborted.", __FILE__, __LINE__);

			if(!iList)
				qFatal("Fatal error at %s@%d. Program aborted.", __FILE__, __LINE__);

			for(int iter = 0; iter < size(); ++iter)
			{
				mzList->append(at(iter)->m_mz);
				iList->append(at(iter)->m_i);
			}

			return;
		}


	void
		MassSpectrum::toVectors(QVector<double> *mzVector,
				QVector<double> *iVector) const
		{
			if(!mzVector || !iVector)
				qFatal("Fatal error at %s@%d. Program aborted.", __FILE__, __LINE__);

			for(int iter = 0; iter < size(); ++iter)
			{
				mzVector->append(at(iter)->m_mz);
				iVector->append(at(iter)->m_i);
			}

			return;
		}


	void
		MassSpectrum::toMap(QMap<double, double> *map) const
		{
			if(!map)
				qFatal("Fatal error at %s@%d. Program aborted.", __FILE__, __LINE__);

			for(int iter = 0; iter < size(); ++iter)
			{
				map->insert(at(iter)->m_mz, at(iter)->m_i);
			}

			return;
		}


	double
		MassSpectrum::intensitySum()
		{
			// We must return the sum of all the intensities.
			double iSum = 0;

			for(int iter = 0; iter < size(); ++iter)
			{
				iSum += at(iter)->m_i;
			}

			return iSum;
		}


	QString *
		MassSpectrum::asText()
		{
			QList<double> mzLista = mzList();
			QList<double> iLista  = iList();

			QString *text = new QString();

			for(int iter = 0; iter < mzLista.size(); ++iter)
			{
				text->append(QString("%1 %2\n")
						.arg(mzLista.at(iter), 0, 'f', 10)
						.arg(iLista.at(iter), 0, 'f', 10));
			}

			return text;
		}


	void
		MassSpectrum::textToClipboard()
		{
			QClipboard *clipboard = QApplication::clipboard();
			QString *   text      = asText();
			if(text)
			{
				clipboard->setText(*text);
				delete text;
			}
		}


	MassSpectrum *
		MassSpectrum::createFromBase64EncByteArrays(const QByteArray *mzByteArray,
				const QByteArray *iByteArray,
				int               compressionType)
		{
			if(mzByteArray == nullptr || iByteArray == nullptr)
				qFatal("Fatal error at %s@%d. Program aborted.", __FILE__, __LINE__);

			// Start by crunching the mz data.

			QByteArray *p_decByteArray = nullptr;

			// Perform the base64 decoding right away.
			bool ok =
				Base64EncDec::decode(mzByteArray, &p_decByteArray, compressionType);
			if(!ok)
				qFatal("Fatal error at %s@%d. Program aborted.", __FILE__, __LINE__);

			// Now that the decode has been performed, let's iterate in the byte
			// array and read these data as double (8 byte-long words).

			// We'll need to iterate in a byte array and we need to know how many
			// bytes are required to make a double (8 bytes now, 20160922).
			int doubleSize = sizeof(double);

			// We'll store the iterated values in here:
			double iterValue;

			QList<double> mzList;

			for(int iter = 0; iter < p_decByteArray->size(); iter += doubleSize)
			{
				QByteArray midByteArray = p_decByteArray->mid(iter, doubleSize);

				iterValue = *reinterpret_cast<double *>(midByteArray.data());
				mzList.append(iterValue);
			}

			// qDebug() << __FILE__ << __LINE__
			// << "Done creating mzList of items: " << mzList.size();

			// p_decByteArray now point to heap-allocated data that we need to free.
			// Then set it to nullptr and reuse that pointer by providing the
			// function with a pointer to it.
			delete p_decByteArray;
			p_decByteArray = nullptr;


			// And now do the same for the i data.

			// Perform the base64 decoding right away.
			ok = Base64EncDec::decode(iByteArray, &p_decByteArray, compressionType);
			if(!ok)
				qFatal("Fatal error at %s@%d. Program aborted.", __FILE__, __LINE__);

			// Now that the decode has been performed, let's iterate in the byte
			// array and read these data as double (8 byte-long words).

			QList<double> iList;

			for(int iter = 0; iter < p_decByteArray->size(); iter += doubleSize)
			{
				QByteArray midByteArray = p_decByteArray->mid(iter, doubleSize);

				iterValue = *reinterpret_cast<double *>(midByteArray.data());
				iList.append(iterValue);
			}

			// p_decByteArray now point to heap-allocated data that we need to free.
			delete p_decByteArray;

			// qDebug() << __FILE__ << __LINE__
			// << "Done creating iList of items: " << iList.size();

			// At this point we have all the data required to craft and fill-in a
			// new mass spectrum:

			// Sanity check.
			if(mzList.size() != iList.size())
				qFatal("Fatal error at %s@%d. Program aborted.", __FILE__, __LINE__);

			MassSpectrum *spectrum = new MassSpectrum;
			spectrum->initialize(mzList, iList);

			// qDebug() << __FILE__ << __LINE__
			// << "Allocated new spectrum of " << mzList.size() << "peaks";

			return spectrum;
		}


	MassSpectrum *
		MassSpectrum::createFromBase64EncByteArrays(const QByteArray *mzByteArray,
				const QByteArray *iByteArray,
				int               compressionType,
				double mzStart, double mzEnd)
		{
			if(mzByteArray == nullptr || iByteArray == nullptr)
				qFatal("Fatal error at %s@%d. Program aborted.", __FILE__, __LINE__);

			// Start by crunching the mz data.

			QByteArray *p_decByteArray = nullptr;

			// Perform the base64 decoding right away.
			bool ok =
				Base64EncDec::decode(mzByteArray, &p_decByteArray, compressionType);
			if(!ok)
				qFatal("Fatal error at %s@%d. Program aborted.", __FILE__, __LINE__);

			// We can delete the source byte array now to free precious memory.
			delete mzByteArray;

			// Now that the decode has been performed, let's iterate in the byte
			// array and read these data as double (8 byte-long words).

			// We'll need to iterate in a byte array and we need to know how many
			// bytes are required to make a double (8 bytes now, 20160922).
			int doubleSize = sizeof(double);

			// We'll store the iterated values in here:
			double iterValue;

			QList<double> mzList;

			for(int iter = 0; iter < p_decByteArray->size(); iter += doubleSize)
			{
				QByteArray midByteArray = p_decByteArray->mid(iter, doubleSize);

				iterValue = *reinterpret_cast<double *>(midByteArray.data());
				mzList.append(iterValue);
			}

			// qDebug() << __FILE__ << __LINE__
			// << "Done creating mzList of items: " << mzList.size();

			// p_decByteArray now point to heap-allocated data that we need to free.
			// Then set it to nullptr and reuse that pointer by providing the
			// function with a pointer to it.
			delete p_decByteArray;
			p_decByteArray = nullptr;

			// And now do the same for the i data.

			// Perform the base64 decoding right away.
			ok = Base64EncDec::decode(iByteArray, &p_decByteArray, compressionType);
			if(!ok)
				qFatal("Fatal error at %s@%d. Program aborted.", __FILE__, __LINE__);

			// We can delete the source byte array now to free precious memory.
			delete iByteArray;

			// Now that the decode has been performed, let's iterate in the byte
			// array and read these data as double (8 byte-long words).

			QList<double> iList;

			for(int iter = 0; iter < p_decByteArray->size(); iter += doubleSize)
			{
				QByteArray midByteArray = p_decByteArray->mid(iter, doubleSize);

				iterValue = *reinterpret_cast<double *>(midByteArray.data());
				iList.append(iterValue);
			}

			// p_decByteArray now point to heap-allocated data that we need to free.
			delete p_decByteArray;

			// qDebug() << __FILE__ << __LINE__
			// << "Done creating iList of items: " << iList.size();

			// At this point we have all the data required to craft and fill-in a
			// new mass spectrum:

			// Sanity check.
			if(mzList.size() != iList.size())
				qFatal("Fatal error at %s@%d. Program aborted.", __FILE__, __LINE__);

			MassSpectrum *spectrum = new MassSpectrum;
			spectrum->initialize(mzList, iList, mzStart, mzEnd);

			// qDebug() << __FILE__ << __LINE__
			// << "Allocated new spectrum of " << mzList.size() << "peaks";

			return spectrum;
		}


	int
		MassSpectrum::keepCommonSpectra(QList<MassSpectrum *> &      where,
				const QList<MassSpectrum *> &list)
		{
			// Remove from where all the spectra that are not found in list.

			int removed = 0;

			if(where.isEmpty())
				return removed;

			int idx = where.size() - 1;

			while(idx >= 0)
			{
				if(!list.contains(where.at(idx)))
				{
					where.removeAt(idx);
					removed++;
				}

				idx--;
			}

			return removed;
		}


} // namespace msXpertSuite
