/*   msXpertSuite - mass spectrometry software suite
 *   -----------------------------------------------
 *   Copyright(C) 2016 Filippo Rusconi
 *
 *   http://www.{site}.org
 *
 *   This file is part of the msXpertSuite project.
 *
 *   The msXpertSuite project is the successor of the massXpert project. This
 *   project now includes various independent modules:
 *   - massXpert, the original mass spectrometry environment software (greatly
 *     refactored code);
 *   - mobXpert, a module to process and mine mobility mass spectrometry data;
 *   - viewXpert, a powerful TIC chromatogram/mass spectrum viewer/miner;
 *
 *    This program is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */


#ifndef MASS_SPECTRUM_HPP
#define MASS_SPECTRUM_HPP

/////////////////////// Qt includes
#include <QObject>
#include <QList>

/////////////////////// Local includes
#include "MassPeak.hpp"

namespace msXpertSuite
{

	class MassSpectrumList;


	class MassSpectrum : public QObject, public QList<MassPeak *>
	{
		Q_OBJECT

		private:
			QString m_title;
			int     m_lastModifIdx = 0;
			int     m_lastBinIdx = 0;

		public:
			double m_dt = 0.0;
			double m_rt = 0.0;

			int m_requiredDecimals = -1;

			double m_greatestStep = qSNaN();
			int m_binCount = 0;
			bool m_isBinnable = false;	

			double m_binMinMz = 0;
			double m_binMaxMz = 0;
			double m_mzShift = 0;

			MassSpectrum(const QString &title = QString());

			void initialize(const QList<double> &mzList, const QList<double> &iList);
			void initialize(const QList<double> &mzList, const QList<double> &iList,
					double mzStart, double mzEnd);

			void initialize(const QList<double> &mzList, const QList<double> &iList,
					double driftTime);

			void initialize(const QList<double> &mzList, const QList<double> &iList,
					double driftTime, double mzStart, double mzEnd);

			void initialize(const QVector<double> &mzVector, const QVector<double> &iVector);


			~MassSpectrum();

			void reset();
			void deleteMassPeaks();

			double tic();
			double tic(double mzStart, double mzEnd);


			bool setupBinnability(const MassSpectrumList &massSpectra, bool reset = false);
			double fillInBins(QList<double> *binList);
			bool isBinnable(const QList<double> &binList);
			void setupBins(const QList<double> &binList);
			int whichBinIndex(double mz);

			double setGreatestStep(QList<double> *binList);
			double determineMzShift(const MassSpectrum &massSpectrum);

			int combine(const MassPeak &massPeak);
			int combineBinned(const MassPeak &massPeak);

			int combineSlow(const MassPeak &massPeak);

			int combine(const MassSpectrum &massSpectrum);
			int combineBinned(const MassSpectrum &massSpectrum);
			int combine(const MassSpectrum &massSpectrum, double mzStart, double mzEnd);

			int combine(const MassSpectrumList &massSpectra, bool reset = false);
			int combine(const MassSpectrumList &massSpectra,
					double mzStart, double mzEnd,
					bool reset = false);

			int combine(const QList<double> &mzList, const QList<double> &iList);
			int combine(const QList<double> &mzList, const QList<double> &iList,
					double mzStart, double mzEnd);

			int combineIndices(const MassSpectrumList &massSpectra, int firstIdx,
					int lastIdx);

			int contains(double mz) const;

			QList<double> mzList() const;
			QList<double> iList() const;
			void lists(QList<double> *mzList, QList<double> *iList) const;
			void toVectors(QVector<double> *mzVector, QVector<double> *iVector) const;
			void toMap(QMap<double, double> *map) const;

			double intensitySum();

			QString *asText();
			void     textToClipboard();

			static MassSpectrum *
				createFromBase64EncByteArrays(const QByteArray *mzByteArray,
						const QByteArray *iByteArray,
						int               compressionType);
			static MassSpectrum *createFromBase64EncByteArrays(
					const QByteArray *mzByteArray, const QByteArray *iByteArray,
					int compressionType, double mzStart, double mzEnd);

			static int keepCommonSpectra(QList<MassSpectrum *> &      where,
					const QList<MassSpectrum *> &list);
	};


} // namespace msXpertSuite

#endif /* MASS_SPECTRUM_HPP */
