/*   msXpertSuite - mass spectrometry software suite
 *   -----------------------------------------------
 *   Copyright(C) 2016 Filippo Rusconi
 *
 *   http://www.{site}.org
 *
 *   This file is part of the msXpertSuite project.
 *
 *   The msXpertSuite project is the successor of the massXpert project. This
 *   project now includes various independent modules:
 *   - massXpert, the original mass spectrometry environment software (greatly
 *     refactored code);
 *   - mineXpert, a powerful TIC chromatogram/mass/drift spectrum viewer/miner;
 *
 *    This program is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */


/////////////////////// Qt includes
#include <QApplication> // for the clipboard
#include <QClipboard>
#include <QFile>
#include <QDebug>
#include <qmath.h>


/////////////////////// Local includes
#include "MassSpectrumList.hpp"


namespace msXpertSuite
{


#if 0
	static int fileIndex = 0;
#endif 


	MassSpectrumList::MassSpectrumList(const QString &title)
		: QList{}, m_title{title}
	{
	}


	MassSpectrumList::MassSpectrumList(const QList<MassSpectrum *> &massSpectra)
	{
		for(int iter = 0; iter < massSpectra.size(); iter++)
			append(massSpectra.at(iter));
	}


	MassSpectrumList::~MassSpectrumList()
	{
		// We cannot delete the MassSpectrum instances allocated on the heap here:
		// this class is used to get pointers to MassSpectrum instances that do not
		// belong to this object. The user is responsible for calling freeSpectra()
		// at the might moment (typically from inside a MassSpecDataSet object).
		// qDebug() << __FILE__ << __LINE__
		// << "Entering MassSpectrumList::~MassSpectrumList";
	}


	MassSpectrumList &
		MassSpectrumList::operator=(const MassSpectrumList &massSpectra)
		{
			for(int iter = 0; iter < massSpectra.size(); iter++)
				append(massSpectra.at(iter));

			m_title = massSpectra.m_title;
			return *this;
		}


	MassSpectrumList &
		MassSpectrumList::operator=(const QList<MassSpectrum *> &massSpectra)
		{
			for(int iter = 0; iter < massSpectra.size(); iter++)
				append(massSpectra.at(iter));

			return *this;
		}


	void
		MassSpectrumList::freeSpectra()
		{

			qDebug() << __FILE__ << __LINE__
				<< "Entering MassSpectrumList::freeSpectra";

			// Iterate in the list of spectra and delete each one.
			for(int iter = 0; iter < size(); ++iter)
				delete at(iter);

			// And now the parent class.
			clear();
		}


	void 
		MassSpectrumList::characters(double *minMz, double *maxMz, int *greatestSize,
				QList<double> *stepList, double *greatestStepOfAll) const
		{

			MassSpectrum *massSpectrum = nullptr;

			double finalStep = 0;
			double tempStep = 0;

			int finalSize = 0;
			int tempSize = 0;

			double finalMinMz = std::numeric_limits<double>::max();
			double finalMaxMz = 0;


			for(int iter = 0; iter < size(); ++iter)
			{

				// For each spectrum in the mass spectrum list, we will want to the get
				// list of mz differences (steps) throughout all the spectrum.

				QList<double>spectralStepList;

				massSpectrum = at(iter);

				// Start with spectrum size because if 0, we skip.

				tempSize = massSpectrum->size();
				if(!tempSize)
					continue;

				if(greatestSize != nullptr)
				{
					if(tempSize > finalSize)
						finalSize = tempSize;
				}

				if(greatestStepOfAll != nullptr)
				{
					tempStep = at(iter)->fillInBins(&spectralStepList);
#if 0

					QFile file(QString("/tmp/stepList-%1.txt").arg(fileIndex));

					file.open(QIODevice::WriteOnly);

					QTextStream fileStream(&file);

					for(int iter = 0; iter < spectralStepList.size(); ++iter)
					{
						fileStream << QString("%1 %2\n")
							.arg(spectralStepList.at(iter), 0, 'f', 5);
					}

					fileStream << "Greatest step value:" << tempStep;

					++fileIndex;

					fileStream.flush();
					file.close();

#endif

					if(stepList != nullptr)
						stepList->append(tempStep);

					if(tempStep > finalStep)
						finalStep = tempStep;
				}

				if(maxMz != nullptr)
				{
					double tempMaxMz = massSpectrum->last()->m_mz;
					if(tempMaxMz > finalMaxMz)
						finalMaxMz = tempMaxMz;
				}

				if(minMz != nullptr)
				{
					double tempMinMz = massSpectrum->first()->m_mz;
					if(tempMinMz < finalMinMz)
						finalMinMz = tempMinMz;
				}
			}
			// End of 
			// for(int iter = 0; iter < size(); ++iter)

			if(minMz != nullptr)
				*minMz = finalMinMz;
			if(maxMz != nullptr)
				*maxMz = finalMaxMz;
			if(greatestSize != nullptr)
				*greatestSize = finalSize;
			if(greatestStepOfAll != nullptr)
				*greatestStepOfAll = finalStep;
		}


	bool 
		MassSpectrumList::sameLengthSpectra() const
		{
			int specCount = size();

			double specSizeSum = 0;
			double specSizeAvg = 0;

			for(int iter = 0; iter < specCount; ++iter)
			{
				specSizeSum += at(iter)->size();
			}

			specSizeAvg = specSizeSum / specCount;

			double specSizeVarN = 0;

			for(int iter = 0; iter < specCount; ++iter)
			{
				specSizeVarN += (at(iter)->size() - specSizeAvg) * (at(iter)->size() - specSizeAvg);
			}

			double specSizeVar = specSizeVarN / specCount;

			double specSizeStdDev = sqrt(specSizeVar);

			//QString msg = QString("specCount = %1 - specSizeSum: %2 - specSizeAvg = %3 - specSizeVarN = %4"
					//" - specSizeVar = %5 - specSizeStdDev = %6\n")
				//.arg(specCount)
				//.arg(specSizeSum, 0, 'f', 7)
				//.arg(specSizeAvg, 0, 'f', 7)
				//.arg(specSizeVarN, 0, 'f', 7)
				//.arg(specSizeVar, 0, 'f', 7)
				//.arg(specSizeStdDev, 0, 'f', 7);

			//qDebug() << __FILE__ << __LINE__
				//<< msg;

			if(specSizeStdDev > 0.1)
				return false;

			return true;
		}



	double
		MassSpectrumList::tic()
		{
			// Iterate in each mass spectrum of this mass spectrum list and get the tic.
			// Then add it recursively.

			double totalTic = 0;

			for(int iter = 0; iter < size(); ++iter)
			{
				totalTic += at(iter)->tic();
			}

			return totalTic;
		}


	double
		MassSpectrumList::tic(double mzStart, double mzEnd)
		{
			// Iterate in each mass spectrum of this mass spectrum list and get the
			// tic. Then add it recursively.  Do that only if the mz value is in the
			// ranged passed as parameters.

			double totalTic = 0;

			for(int iter = 0; iter < size(); ++iter)
			{
				totalTic += at(iter)->tic(mzStart, mzEnd);
			}

			return totalTic;
		}


	// This function does not seem on a  large file to benefit from
	// paralallelization (20160921).
	double
		MassSpectrumList::tic(const MassSpectrumList &massSpectrumList)
		{
			double totalTic = 0;

			//#pragma omp parallel for reduction (+:totalTic)
			for(int jter = 0; jter < massSpectrumList.size(); ++jter)
			{
				totalTic += massSpectrumList.at(jter)->tic();
			}

			return totalTic;
		}


	int
		MassSpectrumList::combine(const MassSpectrumList &massSpectra,
				MassSpectrum &          combinedMassSpectrum)
		{
			// We receive a list of spectra and we need to compute a combined spectrum
			// inside the other parameter
			// we got.

			return combinedMassSpectrum.combine(massSpectra);
		}


	MassSpectrum *
		MassSpectrumList::combine()
		{
			// We allocate a new mass spectrum which will hold the data
			// corresponding to the combination of all our mass spectra.

			// qDebug() << __FILE__ << __LINE__
			// << "Starting combination of " << massSpectra.size()
			// << "mass spectra into this mass spectrum";

			MassSpectrum *combinedMassSpec = new MassSpectrum;

			for(int iter = 0; iter < size(); ++iter)
			{
				MassSpectrum *iterMassSpectrum = at(iter);

				for(int jter = 0; jter < iterMassSpectrum->size(); ++jter)
				{
					combinedMassSpec->combine(*(iterMassSpectrum->at(jter)));
				}
			}

			return combinedMassSpec;
		}


	MassSpectrum *
		MassSpectrumList::combine(double mzStart, double mzEnd)
		{
			// We allocate a new mass spectrum which will hold the data
			// corresponding to the combination of all our mass spectra.  However, we
			// need to check each individual MassPeak of the various MassSpectrum
			// instances for their m/z value, because it can only be combined if its
			// value is in between both mzStart and mzEnd arguments to this function.

			MassSpectrum *combinedMassSpec = new MassSpectrum;

			for(int iter = 0; iter < size(); ++iter)
			{
				MassSpectrum *iterMassSpectrum = at(iter);

				for(int jter = 0; jter < iterMassSpectrum->size(); ++jter)
				{
					MassPeak *iterMassPeak = iterMassSpectrum->at(jter);

					if(iterMassPeak->m_mz >= mzStart && iterMassPeak->m_mz <= mzEnd)
						combinedMassSpec->combine(*iterMassPeak);
				}
			}

			return combinedMassSpec;
		}


	int
		MassSpectrumList::keepCommonSpectra(MassSpectrumList &      where,
				const MassSpectrumList &list)
		{
			// Remove from where all the spectra that are not found in list.

			int removed = 0;

			if(where.isEmpty())
				return removed;

			int idx = where.size() - 1;

			while(idx >= 0)
			{
				if(!list.contains(where.at(idx)))
				{
					where.removeAt(idx);
					removed++;
				}

				idx--;
			}

			return removed;
		}


} // namespace msXpertSuite
