/*   msXpertSuite - mass spectrometry software suite
 *   -----------------------------------------------
 *   Copyright(C) 2016 Filippo Rusconi
 *
 *   http://www.{site}.org
 *
 *   This file is part of the msXpertSuite project.
 *
 *   The msXpertSuite project is the successor of the massXpert project. This
 *   project now includes various independent modules:
 *   - massXpert, the original mass spectrometry environment software (greatly
 *     refactored code);
 *   - mobXpert, a module to process and mine mobility mass spectrometry data;
 *   - viewXpert, a powerful TIC chromatogram/mass spectrum viewer/miner;
 *
 *    This program is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */


////////////////////////////// Qt includes
#include <QRegExp>
#include <QFileInfo>
#include <QDataStream>
#include <QDebug>
#include <QRegularExpressionMatch>


////////////////////////////// Local includes
#include "globals.hpp"


namespace msXpertSuite
{


  QChar gConfigKeyStringPrefixDelimiter = QChar('%');

  int ATOM_DEC_PLACES     = 10;
  int OLIGOMER_DEC_PLACES = 5;
  int POLYMER_DEC_PLACES  = 3;
  int PH_PKA_DEC_PLACES   = 2;

  // How many times the whole gaussian/lorentzian curve should span
  // around the centroid mz ratio. 4 means that the peak will span
  // [ mzRatio - (2 * FWHM) --> mzRatio + (2 * FWHM) ]

  int FWHM_PEAK_SPAN_FACTOR = 4;


  // Useful to send <<-based string output to the console, non debugging mode.
  QTextStream &
  qStdOut()
  {
    static QTextStream ts(stdout);
    return ts;
  }


  QString &
  GlobUnspacifyQString(QString &src)
  {
    if(src.isEmpty())
      {
        return src;
      }

    src.remove(QRegExp("\\s+"));

    return src;
  }


  QString
  GlobBinaryRepresentation(int value)
  {
    QString string;
    string = QString("%1").arg(value, 32, 2);

    return string;
  }


  QString
  GlobElideText(const QString &text, int charsLeft, int charsRight,
                const QString &delimitor)
  {
    // We want to elide text. For example, imagine we have text = "that
    // is beautiful stuff", with charsLeft 4 and charsRight 4 and
    // delimitor "...". Then the result would be "that...tuff"

    if(charsLeft < 1 || charsRight < 1)
      {
        return text;
      }

    int size = text.size();

    // If the text string is already too short, no need to elide
    // anything.
    if((charsLeft + charsRight + delimitor.size()) >= size)
      {
        return text;
      }

    QString result = text.left(charsLeft);
    result.append(delimitor);
    result.append(text.right(charsRight));

    return result;
  }



	// Convert a possibly very long string with not newline chars to 
	// a paragraph of given width by inserting newline chars at the 
	// proper place(s).
	QString 
		GlobStanzify(const QString &text, int width)
		{
			QString result = text;

			// First, replace all the existing newline chars with spaces.

			result = result.replace("\n"," ");

			int iter = width;
			
			// Then, iterate in the obtained string and every width characters try to
			// insert a newline character by iterating back to the left and searching
			// for a space.

			for(; iter < result.size(); iter += width)
			{
				// Now iterate in reverse and search for a space where to insert a
				// newline
				
				int jter = iter;

				for(; jter >= 0; --jter)
				{
					if(result.at(jter) == ' ')
					{
						result[jter] = '\n';
						break;
					}
				}
			}

			return result;
		}


  QString
  GlobMzIFieldSeparator(const QString &text)
  {
    // The separator between mz and i might be anything but a numerical
    // character. Let's try to figure this out.
    QRegularExpression regExp(
      "^(\\d*.\\d+)([^\\d.]*)(-?\\d*\\.?\\d*[e-]?\\d*)");
    QRegularExpressionMatch match = regExp.match(text);

    if(match.hasMatch())
      {
        QString separator = match.captured(2);
        return separator;
      }

    return QString();
  }

  int
  GlobKeepInRange(QList<double> &list, double start, double end)
  {
    // This function assumes that the list was sorted.

    int removed = 0;

    if(list.isEmpty())
      {
        return removed;
      }

    int idx = list.size() - 1;

    while(idx >= 0)
      {
        double iterVal = list.at(idx);

        if(iterVal < start || iterVal > end)
          {
            list.removeAt(idx);
            removed++;
          }

        idx--;
      }

    return removed;
  }


  bool
  GlobIsSqlite3File(QString fileName)
  {
    QFileInfo fileInfo(fileName);

    if(!fileInfo.exists())
      {
        return false;
      }

    QFile file(fileName);

    bool result = file.open(QIODevice::ReadOnly);

    if(!result)
      {
        return false;
      }

    QDataStream inStream(&file);
    char        text[16];
    inStream.readRawData(text, 15);

    QString readText(text);

    if(readText == "SQLite format 3")
      {
        return true;
      }

    //     qDebug() << __FILE__ << __LINE__
    //     << "text:" << text;

    return false;
  }


  bool
  GlobIsMzmlFile(QString fileName)
  {
    QFileInfo fileInfo(fileName);

    if(!fileInfo.exists())
      {
        return false;
      }

    QFile file(fileName);

    bool result = file.open(QIODevice::ReadOnly);

    if(!result)
      {
        return false;
      }

    QDataStream inStream(&file);
    char        text[90];
    inStream.readRawData(text, 89);

    QString readText(text);

    if(readText.contains("<indexedmzML xmlns=\"http://psi.hupo.org/ms/mzml\""))
      {
        return true;
      }
    else if(readText.contains("<mzML xmlns=\"http://psi.hupo.org/ms/mzml\""))
      {
        // In the case of Bruker .d (analysis.baf) files, there is not
        // indexedmzMl
        // element, but simply mzML.

        return true;
      }

    // qDebug() << __FILE__ << __LINE__ << "File " << fileName
    //<< "is not an mzML file";

    return false;
  }


	// Fixme: fake, just to have the soft compile.
  bool
  GlobIsXyFile(QString fileName, bool parseFull, int *spectrumCount)
  {
    return true;
  }


  bool
  GlobIsDxFile(QString fileName, bool parseFull, int *spectrumCount)
  {
    // The header of a dx file as produced by Bruker CompassXtractor from .d
    // files (with analysis.baf inside the directory) and from microQTOF
    // instruments is as follows:

    //##TITLE= ANALYSIS.BAF
    //##JCAMP-DX= 6.00
    //##DATA TYPE= MASS SPECTRUM
    //##DATA CLASS= NTUPLES
    //##ORIGIN=  DA_YC
    //##OWNER=
    //##LONGDATE= 2014/06/17 15:22:41.531 +01:00
    //##SOURCE REFERENCE= C:\Users\<name>\directory\<filedir>.d\ANALYSIS.BAF
    //##SAMPLE DESCRIPTION=
    //##DATA PROCESSING= RECALIBRATION,  ,
    //##SPECTROMETER/DATA SYSTEM= microTOFLC Q
    //##NTUPLES= MASS SPECTRUM
    //##VAR_NAME= MASS, INTENSITY, RETENTION TIME
    //##SYMBOL= X, Y, T
    //##SYMBOL= INDEPENDENT, DEPENDENT, INDEPENDENT
    //##VAR_FORM= AFFN, AFFN, AFFN
    //##VAR_DIM= , , 101
    //##UNITS= M/Z, RELATIVE ABUNDANCE, SECONDS
    //##FIRST= 1.158000
    //##LAST= 101.538000

    // Then, each spectrum has a header like this and then the data:

    //##PAGE= T= 1.158000
    //##NPOINTS= 74528
    //##DATA TABLE= (XY..XY), PEAKS
    // 694.499324,35.000000;694.516080,44.000000;... goes on and on in a single
    // line of mz,i pairs.

    int LIMIT_PARSED_SPECTRA_COUNT = 2;

    QFileInfo fileInfo(fileName);

    if(!fileInfo.exists())
      {
        return false;
      }

    QFile file(fileName);

    if(!file.open(QFile::ReadOnly | QFile::Text))
      {
        qDebug() << __FILE__ << __LINE__ << "Failed to open file" << fileName;

        return false;
      }

    QRegularExpression      endOfLine("^\\s+$");
    QRegularExpressionMatch match;

    int spectraRead = 0;

    bool jcampDxLineFound          = false;
    bool dataTablePeaksLineWasLast = false;

    while(!file.atEnd())
      {
        if(!parseFull && spectraRead >= LIMIT_PARSED_SPECTRA_COUNT)
          {
            break;
          }

        QString line = file.readLine();

        // qDebug() << __FILE__ << __LINE__
        //<< "line is: " << line;

        if(line.isEmpty())
          {
            // qDebug() << __FILE__ << __LINE__
            //<< "line is empty.";

            continue;
          }

        // If the line has only a newline character, go on.
        match = endOfLine.match(line);

        if(match.hasMatch())
          {
            // qDebug() << __FILE__ << __LINE__
            //<< "matched the end of line.";

            continue;
          }

        QRegularExpression jcampDxLine("JCAMP-DX");
        match = jcampDxLine.match(line);

        if(match.hasMatch())
          {
            qDebug() << __FILE__ << __LINE__ << "matched the JCAMP-DX line.";

            jcampDxLineFound = true;

            continue;
          }

        QRegularExpression dataTablePeaksLine(
          "^##DATA TABLE= \\(XY..XY\\), PEAKS\\s*$");
        match = dataTablePeaksLine.match(line);

        if(match.hasMatch())
          {

            qDebug() << __FILE__ << __LINE__
                     << "matched the DATA TABLE PEAKS line.";

            dataTablePeaksLineWasLast = true;

            continue;
          }

        if(line.startsWith("##"))
          {
            dataTablePeaksLineWasLast = false;

            continue;
          }

        // If previous line was the data table peaks line, then this one should
        // be the data. Or there is an error.

        // Format of the numerical string:
        // 694.642704,106.000000;694.759649,92.000000;\n

        QRegularExpression realDataLine(
          "(\\d*\\.?\\d*)([\\s,]*)(-?\\d*\\.?\\d*e?\\d*);");
        match = realDataLine.match(line);

        if(!match.hasMatch() && dataTablePeaksLineWasLast)
          {
            qStdOut() << __FILE__ << __LINE__ << "File " << fileName
                      << "is not a dx file";

            return false;
          }
        else
          {
            // qDebug() << __FILE__ << __LINE__
            //<< "matched the realDataLine with strings:"
            //<< match.captured(1) << "and" << match.captured(3);

            ++spectraRead;
          }
      }

    if(!jcampDxLineFound)
      return false;

    if(spectrumCount != nullptr)
      *spectrumCount = spectraRead;

    return true;
  }


	int GlobZeroDecimals(double value)
	{
		// We want to return how many decimals are NULL:
		//
		// 0.11 would return 0 (no empty decimal)
		// 2.001 would return 2
		// 1000.0001254 would return 3
		
		int intPart = static_cast<int>(value);

		double decimalPart = value - intPart;

		int count = -1;

		while(1)
		{
			++count;

			decimalPart *= 10;

			if(decimalPart > 1)
				return count;
		}

		return count;
	}


} // namespace msXpertSuite
