/*   msXpertSuite - mass spectrometry software suite
 *   -----------------------------------------------
 *   Copyright(C) 2016 Filippo Rusconi
 *
 *   http://www.{site}.org
 *
 *   This file is part of the msXpertSuite project.
 *
 *   The msXpertSuite project is the successor of the massXpert project. This
 *   project now includes various independent modules:
 *   - massXpert, the original mass spectrometry environment software (greatly
 *     refactored code);
 *   - mobXpert, a module to process and mine mobility mass spectrometry data;
 *   - viewXpert, a powerful TIC chromatogram/mass spectrum viewer/miner;
 *
 *    This program is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */


#ifndef GLOBALS_HPP
#define GLOBALS_HPP


////////////////////////////// Qt includes
#include <QString>
#include <QList>
#include <QTextStream>


/*****************************
 * Bitwise stuff (from StackOverflow)

 It is sometimes worth using an enum to name the bits:

 enum ThingFlags = {
 ThingMask  = 0x0000,
 ThingFlag0 = 1 << 0,
 ThingFlag1 = 1 << 1,
 ThingError = 1 << 8,
 }

 Then use the names later on. I.e. write

 thingstate |= ThingFlag1;
 thingstate &= ~ThingFlag0;
 if (thing & ThingError) {...}

 to set, clear and test. This way you hide the magic numbers from the rest of
 your code.
 *******************************/


namespace msXpertSuite
{


	enum UserType
	{
		USER_TYPE_USER   = 0,
		USER_TYPE_SYSTEM = 1
	};

	//! Type of mass.
	/*! Can be monoisotopic or average or both.
	*/
	enum MassType
	{
		MASS_NONE = 0 << 1,                //!< No mass
		MASS_MONO = 1 << 1,                //!< Monoisotopic mass
		MASS_AVG  = 2 << 1,                //!< Average mass
		MASS_BOTH = (MASS_MONO | MASS_AVG) //!< Both masses
	};


	//! Type of mass tolerance.
	/*! Can be atomic mass units, percentage or part per million.
	*/
	enum MassToleranceType
	{
		MASS_TOLERANCE_AMU = 1 << 1, //!< atomic mass units
		MASS_TOLERANCE_PCT = 2 << 1, //!< percentage
		MASS_TOLERANCE_PPM = 3 << 1, //!< parts per million
	};


	//! End of the polymer sequence.
	enum PolymerEnd
	{
		END_NONE  = 0 << 1,
		END_LEFT  = 1 << 1,
		END_RIGHT = 2 << 1,
		END_BOTH  = (END_LEFT | END_RIGHT),
	};


	//! Type of capping.
	/*! The sequence may be capped at left end , at right end, at both
	 *  ends or not capped at all.
	 */
	enum CapType
	{
		CAP_NONE  = 0 << 1,
		CAP_LEFT  = 1 << 1,
		CAP_RIGHT = 2 << 1,
		CAP_BOTH  = (CAP_LEFT | CAP_RIGHT),
	};


	//! Monomer chemical entities.
	/*! The monomer chemical entities can be nothing or modifications.
	*/
	enum MonomerChemEnt
	{
		MONOMER_CHEMENT_NONE       = 0 << 1,
		MONOMER_CHEMENT_MODIF      = 1 << 1,
		MONOMER_CHEMENT_CROSS_LINK = 1 << 2,
	};


	//! Polymer chemical entities.
	/*! The polymer chemical entities can be nothing or left end
	 *  modification or right end modification or both ends modifications.
	 */
	enum PolymerChemEnt
	{
		POLYMER_CHEMENT_NONE                  = 1 << 0,
		POLYMER_CHEMENT_LEFT_END_MODIF        = 1 << 1,
		POLYMER_CHEMENT_FORCE_LEFT_END_MODIF  = 1 << 2,
		POLYMER_CHEMENT_RIGHT_END_MODIF       = 1 << 3,
		POLYMER_CHEMENT_FORCE_RIGHT_END_MODIF = 1 << 4,
		POLYMER_CHEMENT_BOTH_END_MODIF =
			(POLYMER_CHEMENT_LEFT_END_MODIF | POLYMER_CHEMENT_RIGHT_END_MODIF),
		POLYMER_CHEMENT_VORCE_BOTH_END_MODIF =
			(POLYMER_CHEMENT_FORCE_LEFT_END_MODIF |
			 POLYMER_CHEMENT_FORCE_RIGHT_END_MODIF)
	};


	enum PeakShapeType
	{
		PEAK_SHAPE_TYPE_GAUSSIAN   = 1 << 0,
		PEAK_SHAPE_TYPE_LORENTZIAN = 1 << 1,
	};


	enum LogType
	{
		LOG_TO_CONSOLE    = 1 << 0,
		LOG_TO_CONSOLE_OVERWRITE = 1 << 1,
		LOG_TO_STATUS_BAR = 1 << 2,
		LOG_TO_BOTH       = (LOG_TO_CONSOLE | LOG_TO_STATUS_BAR),
		LOG_TO_BOTH_OVERWRITE = (LOG_TO_CONSOLE_OVERWRITE | LOG_TO_STATUS_BAR),
	};


	enum DataCompression
	{
		DATA_COMPRESSION_UNSET = -1,
		DATA_COMPRESSION_NONE  = 0,
		DATA_COMPRESSION_ZLIB  = 1,
	};


	enum MassDataKind
	{
		MASS_SPECTRA = 0,
		DRIFT_SPECTRA,
	};


	enum HashAccountData
	{
		HASH_ACCOUNT_SEQUENCE      = 1 << 0,
		HASH_ACCOUNT_MONOMER_MODIF = 1 << 1,
		HASH_ACCOUNT_POLYMER_MODIF = 1 << 2,
	};

	enum PlotAxis
	{
		NONE_AXIS = 1 << 0,
		X_AXIS    = 1 << 1,
		Y_AXIS    = 1 << 2,
		BOTH_AXES = (X_AXIS | Y_AXIS)
	};

	enum MassSpecDataFileFormat
	{

		// The strategy is to put the most complex formats but easily identified on
		// top, so that when iterating in all the value of the enum, it is a quick
		// operation to get to the proper format (see class
		// MassSpecDataFileFormatAnalyzer).

		MASS_SPEC_DATA_FILE_FORMAT_NOT_SET = 0,
		MASS_SPEC_DATA_FILE_FORMAT_SQLITE3,
		MASS_SPEC_DATA_FILE_FORMAT_MZML,
		MASS_SPEC_DATA_FILE_FORMAT_DX,
		MASS_SPEC_DATA_FILE_FORMAT_MS1,
		MASS_SPEC_DATA_FILE_FORMAT_BRUKER_XY,
		MASS_SPEC_DATA_FILE_FORMAT_XY,
		MASS_SPEC_DATA_FILE_FORMAT_LAST,
	};

	enum HistoryLoggingType 
	{
		Never = 0x0000,
		OnSuccess = 1 << 0,
		OnFailure = 1 << 1,
		Always = OnSuccess | OnFailure
	};
	// Q_ENUM(HistoryLoggingType);

	enum ScriptEntityType
	{
		UndefinedEntity,
		StringEntity,
		FunctionEntity,
		ArrayOrListEntity,
		QObjectEntity,
		IntegerEntity,
		DoubleEntity
	};


	extern int ATOM_DEC_PLACES;
	extern int OLIGOMER_DEC_PLACES;
	extern int POLYMER_DEC_PLACES;
	extern int PH_PKA_DEC_PLACES;

	// How many times the whole gaussian/lorentzian curve should span
	// around the centroid mz ratio. 4 means that the peak will span
	// [ mzRatio - (2 * FWHM) --> mzRatio + (2 * FWHM) ]

	extern int FWHM_PEAK_SPAN_FACTOR;

	extern QChar gConfigKeyStringPrefixDelimiter;

	enum LogTextColor{Comment, Success, Undefined, Failure, JsInput};


	QTextStream &
		qStdOut();

	QString &
		GlobUnspacifyQString(QString &);

	QString 
		GlobBinaryRepresentation(int);

	QString 
		GlobElideText(const QString &, int = 4, int = 4,
				const QString & = "...");

	QString 
		GlobStanzify(const QString &text, int width);

	QString GlobMzIFieldSeparator(const QString &text);

	int 
		GlobKeepInRange(QList<double> &list, double start, double end);

	bool 
		GlobIsSqlite3File(QString fileName);

	bool 
		GlobIsMzmlFile(QString fileName);

	bool 
		GlobIsXyFile(QString fileName, bool parseFull = false,
				int *lineCount = nullptr);

	bool 
		GlobIsDxFile(QString fileName, bool parseFull = false,
				int *spectrumCount = nullptr);

	bool 
		GlobIsMs1File(QString fileName, bool parseFull = false,
				int *lineCount = nullptr);

	int 
		GlobZeroDecimals(double value);

} // namespace msXpertSuite


#endif // GLOBALS_HPP
