/*   msXpertSuite - mass spectrometry software suite
 *   -----------------------------------------------
 *   Copyright(C) 2016 Filippo Rusconi
 *
 *   http://www.{site}.org
 *
 *   This file is part of the msXpertSuite project.
 *
 *   The msXpertSuite project is the successor of the massXpert project. This
 *   project now includes various independent modules:
 *   - massXpert, the original mass spectrometry environment software (greatly
 *     refactored code);
 *   - mineXpert, a powerful TIC chromatogram/mass/drift spectrum viewer/miner;
 *
 *    This program is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */


#ifndef SPECTRUM_CALCULATION_DLG_HPP
#define SPECTRUM_CALCULATION_DLG_HPP


/////////////////////// Qt includes
#include <QDialog>
#include <QTextDocument>


/////////////////////// Local includes
#include "ui_SpectrumCalculationDlg.h"
#include "PolChemDef.hpp"
#include "Atom.hpp"
#include "AtomCount.hpp"
#include "Formula.hpp"
#include "PeakShape.hpp"
#include "globals.hpp"
#include "OligomerList.hpp"


namespace msXpertSuite
{

  enum ValidationError
  {
    VALIDATION_ERRORS_NONE       = 0x0000,
    VALIDATION_FORMULA_ERRORS    = 1 << 0,
    VALIDATION_RESOLUTION_ERRORS = 1 << 1,
    VALIDATION_FWHM_ERRORS       = 1 << 2,
  };

  enum SpectrumCalculationMode
  {
    SPECTRUM_CALCULATION_MODE_NONE     = 0x0000,
    SPECTRUM_CALCULATION_MODE_CLUSTER  = 1 << 0,
    SPECTRUM_CALCULATION_MODE_SPECTRUM = 1 << 1,
  };


  class SpectrumCalculationDlg : public QDialog
  {
    Q_OBJECT

    private:
    Ui::SpectrumCalculationDlg m_ui;

    SpectrumCalculationMode m_mode;

    const QList<Atom *> &m_atomList;

    OligomerList *mp_oligomerList;

    int m_validationErrors;

    QString m_filePath;
    Formula m_formula;

    QString m_configSettingsFilePath;

    double m_minimumProbability;
    int    m_maximumPeaks;

    double m_resolution;

    PeakShapeConfig m_config;

    int m_charge;

    bool m_withMonoMass;
    bool m_withCluster;

    const PolChemDef &m_polChemDef;

    QString m_resultsString;

    // The point list that will hold the sum of all the
    // gaussians/lorentzian curves obtained for the single peak shapes
    // of the spectrum.

    QList<QPointF *> m_patternPointList;

    double m_mono;
    double m_avg;

    bool m_aborted;

    void closeEvent(QCloseEvent *event);

    void setupDialog();

    bool fetchFormulaMass(double * = 0, double * = 0);
    bool fetchMzRatio(double * = 0, double * = 0);

    bool fetchValidateInputData();
    bool fetchValidateInputDataIsotopicClusterMode();
    bool fetchValidateInputDataSpectrumMode();

    void setInErrorStatus(const QString &);

    bool updateMzRatio();
    bool updateIncrement();

    void debugPutStdErrBitset(QString /*file*/, int /*line*/, int = 0 /*value*/,
                              const QString & = QString() /*descString*/);

    void executeCluster();
    void executeSpectrum();
    void executeSpectrumWithCluster(QList<QList<QPointF *> *> *,
                                    QHash<QList<QPointF *> *, double> *);
    void executeSpectrumWithoutCluster(QList<QList<QPointF *> *> *,
                                       QHash<QList<QPointF *> *, double> *);

    private slots:
    void massTypeRadioButtonToggled(bool);
    void isotopicClusterCheckBoxToggled(bool);

    void formulaEdited(const QString &);
    void chargeChanged(int);
    void pointsChanged(int);
    void resolutionChanged(int);
    void fwhmEdited(const QString &);

    void execute();
    void abort();
    void outputFile();

    public:
    SpectrumCalculationDlg(QWidget *      parent,
                           const QString &configSettingsFilePath,
                           const QList<Atom *> &, SpectrumCalculationMode);

    ~SpectrumCalculationDlg();

    void setOligomerList(OligomerList *);

    signals:
    void spectrumCalculationAborted();

    public slots:
    void spectrumCalculationProgressValueChanged(int);
    void spectrumCalculationMessageChanged(QString);
  };

} // namespace msXpertSuite


#endif // SPECTRUM_CALCULATION_DLG_HPP
