/*   msXpertSuite - mass spectrometry software suite
 *   -----------------------------------------------
 *   Copyright(C) 2016 Filippo Rusconi
 *
 *   http://www.{site}.org
 *
 *   This file is part of the msXpertSuite project.
 *
 *   The msXpertSuite project is the successor of the massXpert project. This
 *   project now includes various independent modules:
 *   - massXpert, the original mass spectrometry environment software (greatly
 *     refactored code);
 *   - mineXpert, a powerful TIC chromatogram/mass/drift spectrum viewer/miner;
 *
 *    This program is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */


#include <qmath.h>
#include <omp.h>


/////////////////////// Qt includes
#include <QApplication>
#include <QMultiMap>


/////////////////////// Local includes
#include "ColorMapWnd.hpp"
#include "MainWindow.hpp"
#include "MassSpecDataFileLoaderSqlite3.hpp"
#include "MassSpecDataFileLoaderPwiz.hpp"


namespace msXpertSuite
{

	ColorMapWnd::ColorMapWnd(QWidget *parent)
		: AbstractMultiPlotWnd{parent, "mineXpert" /*m_name*/,
			"Color-map" /*m_desc*/}
	{
		initialize();

		show();
	}

	ColorMapWnd::~ColorMapWnd()
	{
	}


	void
		ColorMapWnd::initialize()
		{
			initializePlotRegion();
		}

	void
		ColorMapWnd::initializePlotRegion()
		{
			// This window, contrary to the other windows, there is not a multigraph
			// plot widget.

			// This window is responsible for the writing of the analysis stanzas to
			// the file. We need to catch the signals.
			// connect(mp_multiGraphPlotWidget,
			//&ColorMapPlotWidget::recordAnalysisStanza,
			// this,
			//&ColorMapWnd::recordAnalysisStanza);
		}


	bool
		ColorMapWnd::focusNextPrevChild(bool next)
		{
			// In this window, we do not want that the TAB key serves to move from
			// widget
			// to widget.
			return false;
		}


	AbstractPlotWidget *
		ColorMapWnd::addPlotWidget(QVector<double> keyVector,
				QVector<double> valVector, const QString &desc,
				MassSpecDataSet *massSpecDataSet)
		{
			// This function is required because it is virtual void = 0 in the base
			// class, but we should never call it in this class.
			qFatal("Fatal error at %s@%d. Program aborted.", __FILE__, __LINE__);
		}


	AbstractPlotWidget *
		ColorMapWnd::initialColorMap(MassSpecDataSet *massSpecDataSet, bool *ok)
		{
			if(massSpecDataSet == nullptr || ok == nullptr)
				qFatal("Fatal error at %s@%d. Program aborted.", __FILE__, __LINE__);

			// There are two kinds of situation:
			//
			// 1. The data are loaded in a streamed manner, that is the whole set of
			// mass spectra from the file is NOT stored in RAM. While loading the
			// data, the loader has dealt with the color map data and crafted data
			// usable for creating the map.
			//
			// 2. The data are fully loaded in RAM and thus we use them to calculate
			// the color map.

			// First off, check if we have drift time data, because if not, then we need
			// not do anything here.

			// Make a list of all the dt values, we'll need that list in the same way,
			// either with streamed or full data.
			QList<double> dtList = massSpecDataSet->m_dtHash.uniqueKeys();
			qSort(dtList.begin(), dtList.end());

			// qDebug() << __FILE__ << __LINE__
			//<< "Number of drift time values:" << dtList.size()
			//<< "from" << dtList.first() << "to" << dtList.last();

			int dtListSize = dtList.size();

			if(dtListSize == 0)
			{
				// This is not intrinsically an error. It could be simply that the data
				// are not for a mobility mass spectrometry experiment.
				*ok = true;
				return nullptr;
			}

			// Make a real hash in the case the data were FULL-loaded, because we
			// cannot use the massSpecDataSet->m_dtHash, as it has already the full
			// data stuff. In the case of a streamed load that m_dtHash is actually
			// used since it would be empty anyway.
			QHash<double, MassSpectrum *> dtSpecMapFullLoad;

			// Now make a placeholder where to store the address of the hash that
			// we'll actually use, either in full or in streamed load.
			QHash<double, MassSpectrum *> *dtSpecHash;

			// Make sure the cancel button is visible.
			mp_statusBarCancelPushButton->setVisible(true);
			QApplication::processEvents();

			MainWindow *mainWindow = static_cast<MainWindow *>(parent());
			QString msgSkeleton = "Computing MZ=DT color map";
			mainWindow->logConsoleMessage(msgSkeleton);

			qDebug() << __FILE__ << __LINE__ << msgSkeleton;

			ColorMapPlotWidget *widget = new ColorMapPlotWidget(
					this, this, m_name, m_desc, massSpecDataSet->m_fileName, massSpecDataSet);

			widget->setMinimumSize(400, 150);
			mp_scrollAreaVBoxLayout->addWidget(widget);

			int mzCellNumber = 0;
			double minMz = 1000000000;
			double maxMz = 0;

			double start = omp_get_wtime();

			// We want to declare/define this variable here
			// because we'll need it later if the loop is interrupted.
			// (see below).
			int parsedDtValues = 0;

			if(massSpecDataSet->m_isStreamed)
			{

				// Since all the data have been streamed already, let's set the number
				// of parsed dt values to the size of the dtList.

				parsedDtValues = dtListSize;

				dtSpecHash = &massSpecDataSet->m_dtHash;

				// Fill in the slots with data that were computed real time while
				// loading the data in streamed mode.
				mzCellNumber = massSpecDataSet->m_mzCellCount;

				minMz = massSpecDataSet->m_mzRangeStart;
				maxMz = massSpecDataSet->m_mzRangeEnd;
			}
			else /* data were loaded fully */
			{
				dtSpecHash = &dtSpecMapFullLoad;

				setupProgressFeedback("Color map calculation: ", 0, dtListSize);

				for(; parsedDtValues < dtListSize; ++parsedDtValues)
				{

					if(m_isOperationCancelled)
					{
						break;
					}

					double dt = dtList.at(parsedDtValues);

					updateProgressFeedback(QString("dt: %1").arg(dt, 0, 'f', 6),
							parsedDtValues);

					// Since were had loaded the full set of data from file,
					// massSpecDataSet->m_dtHash contains all the dt, spectra pairs.
					// Only
					// get the list of spectra for the current dt.
					MassSpectrumList massSpectra = massSpecDataSet->m_dtHash.values(dt);

					int massSpectraSize = massSpectra.size();

					MassSpectrum *combinedMassSpectrum = new MassSpectrum;

					for(int jter = 0; jter < massSpectraSize; ++jter)
					{
						combinedMassSpectrum->m_requiredDecimals = 0;
						combinedMassSpectrum->combine(*massSpectra.at(jter));
					}

					if(combinedMassSpectrum->size() > mzCellNumber)
					{
						mzCellNumber = combinedMassSpectrum->size();
					}

					if(combinedMassSpectrum->first()->m_mz < minMz)
					{
						minMz = combinedMassSpectrum->first()->m_mz;
					}

					if(combinedMassSpectrum->last()->m_mz > maxMz)
					{
						maxMz = combinedMassSpectrum->last()->m_mz;
					}


					// Now put the dt,spectrum pair in the hash that we'll later use to
					// fill-in the colormap.
					dtSpecHash->insert(dt, combinedMassSpectrum);

					// qDebug() << __FILE__ << __LINE__
					// << "Inserted dt value:" << dt << "with spectrum of" <<
					// combinedMassSpectrum->size() << "peaks";
				}

				// End of
				// for(; parsedDtValues < dtListSize; ++parsedDtValues)
			}

			// End of
			// else /* data were loaded fully */

			// Hide the cancel button.
			mp_statusBarCancelPushButton->setVisible(false);

			// We may have gotten here because the operation was cancelled. In any case,
			// reset the boolean value.
			m_isOperationCancelled = false;

			// In some seconds, hide the progress bar and set the label to an empty
			// text.
			QTimer::singleShot(3000, this, [=]() {
					this->updateProgressFeedback("", 0);
					this->mp_statusBarProgressBar->setVisible(false);
					});

			// Now, feed the widget with some data. We only allocate enough space
			// to fit the actual dt values that we could go through (if the operation
			// was cancelled then parsedDtValues is less than dtListSize).

			widget->mp_colorMap->data()->setSize(parsedDtValues, mzCellNumber);
			widget->mp_colorMap->data()->setRange(
					QCPRange(dtList.first(), dtList.last()), QCPRange(minMz, maxMz));
			widget->mp_colorMap->data()->fill(0);

			// Let's see if one color was set as the desired color, otherwise use
			// nothing.
			if(m_wishedColor.isValid())
			{
				// I quit coloring the background of the widget because that was too
				// heavy on the eyes.
				// widget->setBackground(QBrush(m_wishedColor));

				QCPAxis *xAxis = widget->mp_colorMap->keyAxis();

				QPen basePen;
				basePen = xAxis->basePen();
				basePen.setColor(m_wishedColor);

				xAxis->setBasePen(basePen);
				xAxis->setLabelColor(m_wishedColor);
				xAxis->setTickLabelColor(m_wishedColor);

				QCPAxis *yAxis = widget->mp_colorMap->valueAxis();
				yAxis->setBasePen(basePen);
				yAxis->setLabelColor(m_wishedColor);
				yAxis->setTickLabelColor(m_wishedColor);
			}

			// So we now know that we have a m/z amplitude of maxMzValue and that we
			// have maxSpecSize cells.
			// Which means that each cell equiparates to (maxMzValue / maxSpecSize)

			// //double dtPerCell = (dtList.last() - dtList.first()) / dtListSize;
			// double mzPerCell = (maxMz - minMz) / mzCellNumber;

			// qDebug("mzCellNumber: %d \n minMz: %.5f \n maxMz: %.5f \n mzPerCell:
			// %.5f\n",
			// mzCellNumber, minMz, maxMz, mzPerCell);

			// We can reuse the dtList, even if we'll access other data, because the
			// dt values should be the same.
			for(int kter = 0; kter < parsedDtValues; ++kter)
			{
				double dt = dtList.at(kter);

				MassSpectrum *massSpectrum = dtSpecHash->value(dt);

				for(int lter = 0; lter < massSpectrum->size(); ++lter)
				{
					double mz = massSpectrum->at(lter)->m_mz;
					double i = massSpectrum->at(lter)->m_i;

					// This is the version for which the colormap is doing the
					// calculation
					// itself to define what cell need to be dealt with, depending on
					// the
					// values (dt, mz).
					// There might be something in the cell corresponding to (dt, mz),
					// thus try to get the value and make a sum.
					double prevI = widget->mp_colorMap->data()->data(dt, mz);
					widget->mp_colorMap->data()->setData(dt, mz, prevI + i);

					// This is the version where we do the calculation that allows us to
					// know what cell is the one to be dealt with depending on the
					// values
					// (dt, mz).
					// int dtCell = dt / dtPerCell;
					// int mzCell = mz / mzPerCell;
					// There might be something in the cell corresponding to (dt, mz),
					// thus try to get the value and make a sum.
					// double prevI = widget->mp_colorMap->data()->cell(dtCell, mzCell);
					// widget->mp_colorMap->data()->setCell(dtCell, mzCell, prevI + i);

					// qDebug("setting data: dt=%.5f, mz=%.5f, prevI = %.5f, i=%.5f",
					// dt, mz, prevI, i);
				}
			}

			// End of
			// for(int kter = 0; kter < parsedDtValues; ++kter)

			// At this point we have finished filling-up the color map.

			// The gpThermal is certainly one of the best.
			widget->mp_colorMap->setGradient(QCPColorGradient::gpThermal);
			// widget->mp_colorMap->setGradient(QCPColorGradient::gpSpectrum);
			widget->mp_colorMap->rescaleDataRange(true);
			widget->rescaleAxes();
			widget->replot();

			double end = omp_get_wtime();

			QString msg = QString("Computed MZ=DT color map in %1 s.").arg(end - start);

			// Because only the master thread can update the feedback data, the loop
			// might end while we have not reached an iterationCount corresponding to
			// 100%. This might make the user uncomfortable. So we simulate the 100%
			// achievement:
			updateProgressFeedback(msg, parsedDtValues);

			QTimer::singleShot(3000, this, [=]() {
					this->updateProgressFeedback("", 0);
					this->mp_statusBarProgressBar->setVisible(false);
					});

			mainWindow->logConsoleMessage(msg);
			qDebug() << __FILE__ << __LINE__ << msg;

			// At this point, immediately craft a new HistoryItem instance to append
			// to the widget's History.

			HistoryItem *histItem = new HistoryItem;

			// Because we are documenting the mz = f(dt) colormap, we need two
			// integration ranges: the MZ range and the DT range.

			histItem->newIntegrationRange(IntegrationType::DATA_TO_DT_MZ_DT_RANGE,
					dtList.first(),
					dtList.at(parsedDtValues - 1));

			histItem->newIntegrationRange(IntegrationType::DATA_TO_DT_MZ_MZ_RANGE,
					minMz, maxMz);

			widget->addHistoryItem(histItem);

			// Log this recent history to the console.
			mainWindow->logConsoleMessage(histItem->asText());
			mainWindow->logConsoleMessage(widget->m_history.innermostRangesAsText());

			// The plot widget reacts to the Qt::Key_Delete key by calling a function
			// that emits a set of signals, of which the destroyPlotWidget signal
			// actually indirectly triggers the destruction of the plot widget. This
			// signal is
			// now connected to the main window where a function will call for the
			// destruction of all the target plot widgets, that is all the plot
			// widgets that were created as the result of an integration computation
			// of the plot widget currently being destroyed.

			connect(widget, &AbstractPlotWidget::destroyPlotWidget, mainWindow,
					&MainWindow::destroyAlsoAllTargetPlotWidgets);

			// Ensure that there is a connection between this plot widget and this
			// containing window, such that if there is a requirement to replot, the
			// widget gets the signal.
			connect(this, &ColorMapWnd::replotWithAxisRange, widget,
					&ColorMapPlotWidget::replotWithAxisRange);

			// Make sure we get the signal requiring to redraw the background of the
			// widget, depending on it being focused or not.
			connect(this, &AbstractMultiPlotWnd::redrawPlotBackground, widget,
					&AbstractPlotWidget::redrawPlotBackground);

			// At this point make the connections that will allows triggering the
			// mass spectrum and drift spectrum calculations based on the selections
			// over the colormap.

			connect(static_cast<ColorMapPlotWidget *>(widget),
					&ColorMapPlotWidget::newMassSpectrumByDtAndMz,
					mainWindow->mp_massSpecWnd,
					&MassSpecWnd::newMassSpectrumByDtAndMz);

			connect(static_cast<ColorMapPlotWidget *>(widget),
					&ColorMapPlotWidget::newDriftSpectrumByDtAndMz,
					mainWindow->mp_driftSpecWnd,
					&DriftSpecWnd::newDriftSpectrumByDtAndMz);


			connect(widget, &AbstractPlotWidget::updateStatusBarSignal, this,
					&AbstractMultiPlotWnd::updateStatusBar);

			*ok = true;
			return widget;
		}


	// We are not going to produce any graph, we are integrating the data to a
	// single TIC intensity value.
	void 
		ColorMapWnd::ticIntensity(MassSpecDataSet *massSpecDataSet,
				double *intensity, IntegrationType integrationType, History history)
		{
			if(massSpecDataSet == Q_NULLPTR)
				qFatal("Fatal error at %s@%d. Program aborted.", __FILE__, __LINE__);
			if(intensity == Q_NULLPTR)
				qFatal("Fatal error at %s@%d. Program aborted.", __FILE__, __LINE__);

			// We get a History that informs us about the ranges in the
			// sense that the innermost range for integrationType is by necessity the
			// integration range we work on. Note this is a particular case because we
			// need to get two pairs of values since we are called by signal emission
			// from the ColorMapPlotWidget, where two dimensions are simulatenously
			// taken into account: the DT and the MZ.

			double dtStart;
			double dtEnd;

			double mzStart;
			double mzEnd;

			if(!history.innermostRange(IntegrationType::DATA_TO_DT_MZ_DT_RANGE,
						&dtStart, &dtEnd))
				qFatal("Fatal error at %s@%d. Program aborted.", __FILE__, __LINE__);

			if(!history.innermostRange(IntegrationType::DATA_TO_DT_MZ_MZ_RANGE,
						&mzStart, &mzEnd))
				qFatal("Fatal error at %s@%d. Program aborted.", __FILE__, __LINE__);

			MainWindow *mainWindow = static_cast<MainWindow *>(parent());
			QString msgSkeleton =
				QString("Computing Tic intensity between dt %1 and dt %2"
						"and between m/z %3 and m/z %4")
				.arg(dtStart, 0, 'f', 6)
				.arg(dtEnd, 0, 'f', 6)
				.arg(mzStart, 0, 'f', 6)
				.arg(mzEnd, 0, 'f', 6);
			mainWindow->logConsoleMessage(msgSkeleton);

			// Make sure the cancel button is visible.
			mp_statusBarCancelPushButton->setVisible(true);

			double start = omp_get_wtime();

			if(massSpecDataSet->m_isStreamed)
			{
				streamedTicIntensity(massSpecDataSet, 
						intensity, integrationType, history);
			}
			else // that is, !if(massSpecDataSet->m_isStreamed)
			{
				// RT integration.
				// ===============
				bool rtIntegration = false;
				double rtStart = -1;
				double rtEnd = -1;

				rtIntegration = history.innermostRtRange(&rtStart, &rtEnd);
				if(rtIntegration)
				{
					mainWindow->logConsoleMessage(
							"rt integration accounted for: \n" +
							QString("[%1-%2]\n").arg(rtStart).arg(rtEnd));
				}

				// Get handles to the original data in the parent window.
				MsMultiHash *rtHash = &massSpecDataSet->m_rtHash;

				// Get a list of all the retention times that are stored in the whole
				// file and that are contained in the dt range passed as parameters.

				QList<double> rtList;
				if(rtIntegration)
					rtHash->rangedKeys(rtList, rtStart, rtEnd);
				else
					rtList = rtHash->uniqueKeys();

				qSort(rtList.begin(), rtList.end());

				int rtListSize = rtList.size();

				// qDebug() << __FILE__ << __LINE__
				// << "Number of unique retention time keys: " << rtListSize
				// << "from scan start time "<< rtList.first()
				// << "to scan start time" << rtList.last();

				setupProgressFeedback(msgSkeleton, 1, rtListSize);

				// For each retention time, we'll want to
				// 1. get the corresponding mass spectra
				// 2. only retain the spectra that match the drift time (if applicable)
				// 3. compute a TIC value for the retained spectra that is summed 
				// to all the preceding ones, to finally yield a total Tic intensity
				// value (that actually is simply an integration).

				double localIntensity = 0.0;
				int iterationCount = 0;

#pragma omp parallel for
				for(int iter = 0; iter < rtListSize; ++iter)
				{

					if(m_isOperationCancelled)
						continue;

					double rt = rtList.at(iter);

					MassSpectrumList rtMassSpectra(rtHash->values(rt));

					// We have to remove from all the rtMassSpectra, all the spectra
					// that are not in the drift time range.

					for(int jter = 0; jter < rtMassSpectra.size(); ++jter)
					{
						MassSpectrum *spectrum = rtMassSpectra.at(jter);

						// Old version when MassSpectrum did not have m_rt/m_dt
						// member
						// variables.
						// double dt = dtHash->key(spectrum);

						if(spectrum->m_dt < dtStart || spectrum->m_dt > dtEnd)
						{
							// Remove the mass spectrum from the rtMassSpectra
							// because it does
							// not have the proper dt value.
							rtMassSpectra.removeAt(jter);
							--jter;
							continue;
						}
					}

					// At this point, the rtMassSpectra only contain spectra that match
					// both the retention time range and the drift time range.

					for(int kter = 0; kter < rtMassSpectra.size(); ++kter)
					{
						double newTic = 0.0;
						newTic = rtMassSpectra.at(kter)->tic(mzStart, mzEnd);

#pragma omp critical
						localIntensity += newTic;
					}
					// End of for(int kter = 0; kter < massSpectra.size(); ++kter)

					QString msg = msgSkeleton +
						QString("-- rt %1 with %2 spectra")
						.arg(rt, 0, 'f', 6)
						.arg(rtMassSpectra.size());

#pragma omp critical
					{
						++iterationCount;

						// The label and progress bar widgets were setup by the master
						// thread, so we can only modify them (actually emitting signals)
						// from that same master thread:

						if(omp_get_thread_num() == 0)
							updateProgressFeedback(msg, iterationCount);
					}
				}
				// End of
				// #pragma omp parallel for
				// for(int iter = 0; iter < rtList.size(); ++iter)

				// Finally, update the value passed to the function:
				*intensity = localIntensity;
			}
			// End of
			// else // that is, !if(massSpecDataSet->m_isStreamed)

			double end = omp_get_wtime();

			double duration = end - start;

			QString msg = QString("Took %1 s to compute TIC intensity: %2\n")
				.arg(duration)
				.arg(*intensity, 0, 'g', 3);

			// Show 100% completion of the work.
			setupProgressFeedback(msg, 0, 100, 100);

			// Hide the cancel button.
			mp_statusBarCancelPushButton->setVisible(false);

			mainWindow->logConsoleMessage(msg);

			// qDebug() << __FILE__ << __LINE__ << msg;

			// Pretty quickly hide the progress bar and set the label to an empty
			// text so that the user can see the TIC I value on the statusbar.

			// The idea is that if the computation is very quick, then the message
			// should vanish almost immediately. Otherwise let it there longer to show
			// to the user that the computation is computing-intensive.

			// For every second of integration, show the message 20 ms. But do not
			// exceed 3000 ms.
			int timeOut = duration * 20; 
			if(timeOut > 3000)
				timeOut = 3000;

			QTimer::singleShot(timeOut, this, [=]() {
					this->updateProgressFeedback("", 0);
					this->mp_statusBarProgressBar->setVisible(false);
					});

			// We may have gotten here because the operation was cancelled. In any
			// case, reset the boolean value.

			m_isOperationCancelled = false;
		}


	void 
		ColorMapWnd::streamedTicIntensity(MassSpecDataSet *massSpecDataSet,
				double *intensity, IntegrationType integrationType, History history)
		{
			if(intensity == Q_NULLPTR)
				qFatal("Fatal error at %s@%d. Program aborted.",
						__FILE__, __LINE__);

			// Depending on the format of the data file from which the data were
			// loaded in stream mode, it will be possible or not to perform the
			// required integration.

			if(massSpecDataSet->m_fileFormat ==
					MassSpecDataFileFormat::MASS_SPEC_DATA_FILE_FORMAT_SQLITE3)
			{
				MassSpecDataFileLoaderSqlite3 loader(massSpecDataSet->m_fileName);

				// Make sure we connect the feedback signal/slot
				connect(&loader,
						&MassSpecDataFileLoaderSqlite3::setupProgressFeedbackSignal, this,
						&AbstractMultiPlotWnd::setupProgressFeedback);

				connect(&loader,
						&MassSpecDataFileLoaderSqlite3::updateProgressFeedbackSignal, this,
						&AbstractMultiPlotWnd::updateProgressFeedback);

				connect(this, &AbstractMultiPlotWnd::cancelOperationSignal, &loader,
						&MassSpecDataFileLoaderSqlite3::cancelOperation);

				int res = loader.streamedTicIntensity(history, intensity);

				if(res == -1)
					qFatal("Fatal error at %s@%d. Program aborted.", __FILE__, __LINE__);
			}
			else if(massSpecDataSet->m_fileFormat ==
					MassSpecDataFileFormat::MASS_SPEC_DATA_FILE_FORMAT_MZML)
			{
				MassSpecDataFileLoaderPwiz loader(massSpecDataSet->m_fileName);

				// Make sure we connect the feedback signal/slot
				connect(&loader,
						&MassSpecDataFileLoaderPwiz::setupProgressFeedbackSignal, this,
						&AbstractMultiPlotWnd::setupProgressFeedback);

				connect(&loader,
						&MassSpecDataFileLoaderPwiz::updateProgressFeedbackSignal, this,
						&AbstractMultiPlotWnd::updateProgressFeedback);

				connect(this, &AbstractMultiPlotWnd::cancelOperationSignal,
						&loader, &MassSpecDataFileLoaderPwiz::cancelOperation);

				int res = loader.streamedTicIntensity(history, intensity);

				if(res == -1)
					qFatal("Fatal error at %s@%d. Program aborted.", __FILE__, __LINE__);
			}
			else
			{
				QMessageBox msgBox;
				msgBox.setText(
						"The operation cannot be carried over because the "
						"data were loaded in stream and not from a usable file.");

				msgBox.setStandardButtons(QMessageBox::Ok);
				msgBox.exec();

				// Hide the cancel button.
				mp_statusBarCancelPushButton->setVisible(false);

				return;
			}
		}


} // namespace msxpertsuite
